from exasol_integration_test_docker_environment.lib.base.base_task import BaseTask
from exasol_integration_test_docker_environment.lib.docker.images.push.docker_image_push_task import DockerPushImageTask
from exasol_integration_test_docker_environment.lib.docker.images.task_creator_from_build_tasks import \
    TaskCreatorFromBuildTasks


class PushTaskCreatorFromBuildTasks(TaskCreatorFromBuildTasks):

    def __init__(self, task: BaseTask):
        self.task = task

    def create_task_with_required_tasks(self, build_task, required_task_info):
        push_task = \
            self.task.create_child_task_with_common_params(
                DockerPushImageTask,
                image_name=build_task.image_name,
                required_task_info=required_task_info)
        return push_task



from typing import List

from exasol_integration_test_docker_environment.lib.config.log_config import WriteLogFilesToConsole
from exasol_integration_test_docker_environment.lib.logging.abstract_log_handler import AbstractLogHandler


class ContainerLogHandler(AbstractLogHandler):

    def __init__(self, log_file_path, logger, description: str):
        super().__init__(log_file_path, logger)
        self.db_container_name = description

    def handle_log_line(self, log_line, error: bool = False):
        self._log_file.write(log_line + "\n")
        self._complete_log.append(log_line)

    def get_complete_log(self) -> List[str]:
        return self._complete_log

    def finish(self):
        if self._log_config.write_log_files_to_console == WriteLogFilesToConsole.all:
            self._logger.info("Log %s\n%s", self.db_container_name,
                              "\n".join(self._complete_log))



from pathlib import Path
from typing import List, Generator

import luigi
from luigi import Task

from exasol_integration_test_docker_environment.lib.base.stoppable_base_task import StoppableBaseTask
from exasol_integration_test_docker_environment.lib.base.task_dependency import DependencyType, DependencyState, \
    TaskDependency, TaskDescription


class DependencyLoggerBaseTask(StoppableBaseTask):

    def _get_dependencies_path_for_job(self):
        return Path(super()._get_output_path_for_job(), "dependencies")

    def _get_dependencies_path(self):
        path = Path(self._get_dependencies_path_for_job(), self.task_id)
        path.mkdir(parents=True, exist_ok=True)
        return path

    def _get_dependencies_requires_path(self):
        return Path(self._get_dependencies_path(), "requires")

    def _get_dependencies_dynamic_path(self):
        return Path(self._get_dependencies_path(), "dynamic")

    def handle_requires_value(self, tasks):
        dependency_path = self._get_dependencies_requires_path()
        if not dependency_path.exists():
            with dependency_path.open("w") as dependencies_file:
                self.write_dependency(
                    dependencies_file=dependencies_file,
                    dependency_type=DependencyType.requires,
                    dependency_state=DependencyState.requested,
                    index=0, value=tasks)
                return tasks
        else:
            return tasks

    def handle_requires_generator(self, tasks):
        dependency_path = self._get_dependencies_requires_path()
        if not dependency_path.exists():
            with dependency_path.open("w") as dependencies_file:
                result = list(self.write_dependencies_for_generator(
                    dependencies_file=dependencies_file,
                    task_generator=tasks,
                    dependency_type=DependencyType.requires))
                return result
        else:
            return tasks

    def write_dependencies_for_generator(self,
                                         dependencies_file,
                                         task_generator,
                                         dependency_type: DependencyType):
        index = 0
        try:
            element = next(task_generator)
            while True:
                self.write_dependency(
                    dependencies_file=dependencies_file,
                    dependency_type=dependency_type,
                    dependency_state=DependencyState.requested,
                    index=index,
                    value=element)
                result = yield element
                element = task_generator.send(result)
                index += 1
        except StopIteration:
            pass

    def write_dependency(self,
                         dependencies_file,
                         dependency_type: DependencyType,
                         dependency_state: DependencyState,
                         index: int,
                         value):
        for task in self.flatten_tasks(value):
            dependency = TaskDependency(source=self.get_task_description(),
                                        target=task.get_task_description(),
                                        type=dependency_type,
                                        index=index,
                                        state=dependency_state)
            dependencies_file.write(f"{dependency.to_json()}")
            dependencies_file.write("\n")

    def requires(self):
        tasks = super().requires()
        if tasks is not None:
            if isinstance(tasks, (Task, list, tuple, dict)):
                return self.handle_requires_value(tasks)
            else:
                return self.handle_requires_generator(tasks)
        else:
            return []

    def run(self):
        dependency_path = self._get_dependencies_dynamic_path()
        with dependency_path.open("a") as dependencies_file:
            task_generator = super().run()
            if task_generator is not None:
                yield from self.write_dependencies_for_generator(
                    dependencies_file=dependencies_file,
                    task_generator=task_generator,
                    dependency_type=DependencyType.dynamic)

    def get_task_description(self) -> TaskDescription:
        return TaskDescription(id=self.task_id, representation=str(self))

    def flatten_tasks(self, generator: Generator) -> List["DependencyLoggerBaseTask"]:
        return [task for task in luigi.task.flatten(generator)
                if isinstance(task, DependencyLoggerBaseTask)]



import shutil

import luigi

from exasol_integration_test_docker_environment.lib.api.common import generate_root_task
from exasol_integration_test_docker_environment.lib.docker.images.clean.clean_images import CleanImagesStartingWith


def clean(docker_repository_name):
    task = generate_root_task(task_class=CleanImagesStartingWith, starts_with_pattern=docker_repository_name)
    luigi.build([task], workers=1, local_scheduler=True, log_level="INFO")
    if task._get_tmp_path_for_job().exists():
        shutil.rmtree(str(task._get_tmp_path_for_job()))



import ssl
from time import sleep
from urllib.parse import quote_plus
from xmlrpc.client import ServerProxy

import luigi

from exasol_integration_test_docker_environment.lib.base.dependency_logger_base_task import DependencyLoggerBaseTask
from exasol_integration_test_docker_environment.lib.base.json_pickle_parameter import JsonPickleParameter
from exasol_integration_test_docker_environment.lib.data.database_credentials import DatabaseCredentialsParameter
from exasol_integration_test_docker_environment.lib.data.database_info import DatabaseInfo
from exasol_integration_test_docker_environment.lib.data.docker_network_info import DockerNetworkInfo
from exasol_integration_test_docker_environment.lib.test_environment.parameter.external_test_environment_parameter import \
    ExternalDatabaseXMLRPCParameter, ExternalDatabaseHostParameter


class SetupExternalDatabaseHost(DependencyLoggerBaseTask,
                                ExternalDatabaseXMLRPCParameter,
                                ExternalDatabaseHostParameter,
                                DatabaseCredentialsParameter):
    environment_name = luigi.Parameter()
    network_info = JsonPickleParameter(DockerNetworkInfo, significant=False)  # type: DockerNetworkInfo
    attempt = luigi.IntParameter(1)

    def run_task(self):
        database_host = self.external_exasol_db_host
        if self.external_exasol_db_host == "localhost" or \
                self.external_exasol_db_host == "127.0.01":
            database_host = self.network_info.gateway
        self.setup_database()
        database_info = DatabaseInfo(host=database_host,
                                     db_port=self.external_exasol_db_port,
                                     bucketfs_port=self.external_exasol_bucketfs_port,
                                     reused=False)
        self.return_object(database_info)

    def setup_database(self):
        if self.external_exasol_xmlrpc_host is not None:
            # TODO add option to use unverified ssl
            cluster = self.get_xml_rpc_object()
            self.start_database(cluster)
            cluster.bfsdefault.editBucketFS({'http_port': int(self.external_exasol_bucketfs_port)})
            try:
                cluster.bfsdefault.addBucket({'bucket_name': 'myudfs', 'public_bucket': True,
                                              'read_password': self.bucketfs_write_password,
                                              'write_password': self.bucketfs_write_password})
            except Exception as e:
                self.logger.info(e)
            try:
                cluster.bfsdefault.addBucket({'bucket_name': 'jdbc_adapter', 'public_bucket': True,
                                              'read_password': self.bucketfs_write_password,
                                              'write_password': self.bucketfs_write_password})
            except Exception as e:
                self.logger.info(e)

    def get_xml_rpc_object(self, object_name: str = ""):
        uri = 'https://{user}:{password}@{host}:{port}/{cluster_name}/{object_name}'.format(
            user=quote_plus(self.external_exasol_xmlrpc_user),
            password=quote_plus(self.external_exasol_xmlrpc_password),
            host=self.external_exasol_xmlrpc_host,
            port=self.external_exasol_xmlrpc_port,
            cluster_name=self.external_exasol_xmlrpc_cluster_name,
            object_name=object_name
        )
        server = ServerProxy(uri, context=ssl._create_unverified_context())
        return server

    def start_database(self, cluster: ServerProxy):
        storage = self.get_xml_rpc_object("storage")
        # wait until all nodes are online
        self.logger.info('Waiting until all nodes are online')
        all_nodes_online = False
        while not all_nodes_online:
            all_nodes_online = True
            for nodeName in cluster.getNodeList():
                node_state = self.get_xml_rpc_object(nodeName).getNodeState()
                if node_state['status'] != 'Running':
                    all_nodes_online = False
                    break
            sleep(5)
        self.logger.info('All nodes are online now')

        # start EXAStorage
        if not storage.serviceIsOnline():
            if storage.startEXAStorage() != 'OK':
                self.logger.info('EXAStorage has been started successfully')
            else:
                raise Exception('Not able startup EXAStorage!\n')
        elif storage.serviceIsOnline():
            self.logger.info('EXAStorage already online; continuing startup process')

        # triggering database startup
        for databaseName in cluster.getDatabaseList():
            database = self.get_xml_rpc_object('/db_' + quote_plus(databaseName))
            if not database.runningDatabase():
                self.logger.info('Starting database instance %s' % databaseName)
                database.startDatabase()
            else:
                self.logger.info('Database instance %s already running' % databaseName)



import luigi

from exasol_integration_test_docker_environment.cli.options.system_options import DEFAULT_OUTPUT_DIRECTORY


class build_config(luigi.Config):
    force_pull = luigi.BoolParameter(False)
    force_load = luigi.BoolParameter(False)
    force_rebuild = luigi.BoolParameter(False)
    force_rebuild_from = luigi.ListParameter([])
    log_build_context_content = luigi.BoolParameter(False)
    # keep_build_context = luigi.BoolParameter(False)
    temporary_base_directory = luigi.OptionalParameter(None)
    output_directory = luigi.Parameter(DEFAULT_OUTPUT_DIRECTORY)
    cache_directory = luigi.OptionalParameter("")
    build_name = luigi.OptionalParameter("")



import threading
import time
from datetime import datetime, timedelta

from exasol_integration_test_docker_environment.lib.config.log_config import log_config


class StillRunningLoggerThread(threading.Thread):
    def __init__(self, still_running_logger: "StillRunningLogger"):
        super().__init__()
        self.still_running_logger = still_running_logger
        self.finish = False

    def stop(self):
        self.finish = True

    def run(self):
        while not self.finish:
            self.still_running_logger.log()
            time.sleep(1)


class StillRunningLogger:

    def __init__(self, logger, description):
        self._description = description
        self._logger = logger
        self._log_config = log_config()
        self._previous_time = datetime.now()

    def log(self, message=None):

        timedelta_between = timedelta(seconds=self._log_config.seconds_between_still_running_logs)
        if self._previous_time + timedelta_between <= datetime.now():
            if message is None:
                self._logger.info("Still running %s.", self._description)
            else:
                self._logger.info("Still running %s. Message: %s", self._description, message)
            self._previous_time = datetime.now()



import functools
import inspect
import os
import shutil
import subprocess
import tempfile
from pathlib import Path
import shlex
from sys import stderr
from typing import List

from exasol_integration_test_docker_environment.lib.data.environment_info import EnvironmentInfo
from exasol_integration_test_docker_environment.lib.docker import ContextDockerClient
from exasol_integration_test_docker_environment.lib.docker.container.utils import remove_docker_container
from exasol_integration_test_docker_environment.lib.docker.volumes.utils import remove_docker_volumes
from exasol_integration_test_docker_environment.testing.docker_registry import default_docker_repository_name
from exasol_integration_test_docker_environment.testing.exaslct_docker_test_environment import \
    ExaslctDockerTestEnvironment
from exasol_integration_test_docker_environment.testing.spawned_test_environments import SpawnedTestEnvironments
from exasol_integration_test_docker_environment.testing.utils import find_free_ports, check_db_version_from_env


def _cleanup(env_name: str):
    remove_docker_container([f"test_container_{env_name}",
                             f"db_container_{env_name}"])
    remove_docker_volumes([f"db_container_{env_name}_volume"])


def get_class(test_object):
    if test_object is None:
        return None
    # if test_object is a class then simply return it.
    # Otherwise test_object is an instance of a class, then
    # return its class.
    return test_object if inspect.isclass(test_object) else test_object.__class__


def get_test_flavor(test_class):
    if test_class is None:
        return None
    source_file_of_test_object = inspect.getsourcefile(test_class)
    return Path(os.path.realpath(source_file_of_test_object)).parent.joinpath(
        "resources/test-flavor")


class ExaslctTestEnvironment:

    def __init__(self, test_object, executable="./exaslct", clean_images_at_close=True, name=None):
        self.clean_images_at_close = clean_images_at_close
        self.executable = executable
        self.test_object = test_object
        self.test_class = get_class(test_object)
        self.flavor_path = get_test_flavor(self.test_class)
        self.name = name if name else self.test_class.__name__
        self._docker_repository_name = default_docker_repository_name(self.name)
        if "RUN_SLC_TESTS_WITHIN_CONTAINER" in os.environ:
            # We need to put the output directories into the workdir,
            # because only this is shared between the current container and
            # host. Only paths within this shared directory can be mounted
            # to docker container started by exaslct
            temp_dir_prefix_path = Path("./temp_outputs")
            temp_dir_prefix_path.mkdir(exist_ok=True)
            self.temp_dir = tempfile.mkdtemp(dir=temp_dir_prefix_path)
        else:
            self.temp_dir = tempfile.mkdtemp()
        self._update_attributes()

    @property
    def repository_name(self):
        return self._docker_repository_name

    @repository_name.setter
    def repository_name(self, value):
        self._docker_repository_name = value
        self._update_attributes()

    def _update_attributes(self):
        self.flavor_path_argument = f"--flavor-path {self.flavor_path}"
        repository_name = self.repository_name
        self.docker_repository_arguments = f"--source-docker-repository-name {repository_name} " \
                                           f"--target-docker-repository-name {repository_name}"
        self.clean_docker_repository_arguments = f"--docker-repository-name {repository_name}"
        self.output_directory_arguments = f"--output-directory {self.temp_dir}"
        self.task_dependencies_argument = " ".join([f"--task-dependencies-dot-file {self.name}.dot", ])

    def clean_images(self):
        self.run_command(f"{self.executable} clean-flavor-images", clean=True)

    def run_command(self, command: str,
                    use_output_directory: bool = True,
                    use_flavor_path: bool = True,
                    use_docker_repository: bool = True,
                    track_task_dependencies: bool = False,
                    clean: bool = False,
                    capture_output: bool = False):
        if use_output_directory:
            command = f"{command} {self.output_directory_arguments}"
        if track_task_dependencies:
            command = f"{command} {self.task_dependencies_argument}"
        if use_flavor_path:
            command = f"{command} {self.flavor_path_argument}"
        if use_docker_repository and not clean:
            command = f"{command} {self.docker_repository_arguments}"
        if use_docker_repository and clean:
            command = f"{command} {self.clean_docker_repository_arguments}"
        print(file=stderr)
        print(f"command: {command}", file=stderr)
        if capture_output:
            completed_process = subprocess.run(shlex.split(command), stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        else:
            completed_process = subprocess.run(shlex.split(command))
        try:
            completed_process.check_returncode()
        except subprocess.CalledProcessError as e:
            if capture_output:
                print(e.stdout.decode("UTF-8"), file=stderr)
            raise e
        return completed_process

    def close(self):
        try:
            if self.clean_images_at_close:
                self.clean_images()
        except Exception as e:
            print(e, file=stderr)
        try:
            shutil.rmtree(self.temp_dir)
        except Exception as e:
            print(e, file=stderr)

    def spawn_docker_test_environments(self, name: str, additional_parameter: List[str] = None) \
            -> SpawnedTestEnvironments:
        database_port, bucketfs_port = find_free_ports(2)
        on_host_parameter = ExaslctDockerTestEnvironment(
            name=self.name + "_" + name,
            database_host="localhost",
            db_username="sys",
            db_password="exasol",
            bucketfs_username="w",
            bucketfs_password="write",
            database_port=database_port,
            bucketfs_port=bucketfs_port)

        arguments = [
            f"--environment-name {on_host_parameter.name}",
            f"--database-port-forward {on_host_parameter.database_port}",
            f"--bucketfs-port-forward {on_host_parameter.bucketfs_port}",
        ]
        db_version = check_db_version_from_env()
        if db_version:
            arguments.append(f'--docker-db-image-version "{db_version}"')
        if additional_parameter:
            arguments += additional_parameter
        arguments = " ".join(arguments)

        command = f"{self.executable} spawn-test-environment {arguments}"
        completed_process = self.run_command(command, use_flavor_path=False, use_docker_repository=False,
                                             capture_output=True)
        on_host_parameter.completed_process = completed_process
        environment_info_json_path = Path(self.temp_dir,
                                          f"cache/environments/{on_host_parameter.name}/environment_info.json")
        if environment_info_json_path.exists():
            with environment_info_json_path.open() as f:
                environment_info = EnvironmentInfo.from_json(f.read())
                on_host_parameter.environment_info = environment_info
        on_host_parameter.clean_up = functools.partial(_cleanup, on_host_parameter.name)
        if "RUN_SLC_TESTS_WITHIN_CONTAINER" in os.environ:
            slc_test_run_parameter = ExaslctDockerTestEnvironment(
                name=on_host_parameter.name,
                database_host="localhost",
                db_username=on_host_parameter.db_username,
                db_password=on_host_parameter.db_password,
                bucketfs_username=on_host_parameter.bucketfs_username,
                bucketfs_password=on_host_parameter.bucketfs_password,
                database_port=8888,
                bucketfs_port=6583,
                environment_info=on_host_parameter.completed_process,
                completed_process=on_host_parameter.completed_process
            )

            with ContextDockerClient() as docker_client:
                db_container = docker_client.containers.get(f"db_container_{slc_test_run_parameter.name}")
                cloudbuild_network = docker_client.networks.get("cloudbuild")
                cloudbuild_network.connect(db_container)
                db_container.reload()
                slc_test_run_parameter.database_host = \
                    db_container.attrs["NetworkSettings"]["Networks"][cloudbuild_network.name]["IPAddress"]
                return SpawnedTestEnvironments(on_host_parameter, slc_test_run_parameter)
        else:
            return SpawnedTestEnvironments(on_host_parameter, None)




import os


class SymlinkLoopChecker:
    """
    Collects all visited directory inodes in a set and checks for duplicates.
    Can be used to check for symlink loops.
    """
    def __init__(self):
        self._inodes = set()

    def check_and_add(self, directory: str) -> None:
        """
        Checks if given parameter directory has already been added to the internal set of inodes and raises exception
        if found.
        :param directory: The directory which should be checked.
        :raises OSError if the directory already has been passed before.
        """
        stat = os.stat(directory)
        if stat.st_ino > 0 and stat.st_ino in self._inodes:
            raise OSError(
                f"Directory: {directory} contains symlink loops (Symlinks pointing to a parent directory). Please fix!")
        self._inodes.add(stat.st_ino)



import luigi

from exasol_integration_test_docker_environment.lib.data.environment_type import EnvironmentType
from exasol_integration_test_docker_environment.lib.test_environment.parameter.docker_db_test_environment_parameter import \
    DockerDBTestEnvironmentParameter
from exasol_integration_test_docker_environment.lib.test_environment.parameter.external_test_environment_parameter import \
    ExternalDatabaseCredentialsParameter
from exasol_integration_test_docker_environment.lib.test_environment.parameter.general_spawn_test_environment_parameter import \
    GeneralSpawnTestEnvironmentParameter


class SpawnTestEnvironmentParameter(GeneralSpawnTestEnvironmentParameter,
                                    ExternalDatabaseCredentialsParameter,
                                    DockerDBTestEnvironmentParameter):
    environment_type = luigi.EnumParameter(enum=EnvironmentType)



import pathlib

from exasol_integration_test_docker_environment.lib.config.log_config import WriteLogFilesToConsole
from exasol_integration_test_docker_environment.lib.logging.abstract_log_handler import AbstractLogHandler


class CommandLogHandler(AbstractLogHandler):

    def __init__(self, log_file_path: pathlib.Path, logger, description: str):
        super().__init__(log_file_path, logger)
        self._description = description

    def handle_log_line(self, log_line, error: bool = False):
        self._log_file.write(log_line)
        self._complete_log.append(log_line)
        if error:
            self._error_message = log_line

    def finish(self):
        if self._log_config.write_log_files_to_console == WriteLogFilesToConsole.all:
            self._logger.info("Command log for %s \n%s",
                              self._description,
                              "".join(self._complete_log))
        if self._error_message is not None:
            if self._log_config.write_log_files_to_console == WriteLogFilesToConsole.only_error:
                self._logger.error("Command failed %s failed\nCommand Log:\n%s",
                                   self._description,
                                   "\n".join(self._complete_log))
            raise Exception(
                "Error occurred during %s. Received error \"%s\" ."
                "The whole log can be found in %s"
                % (self._description,
                   self._error_message,
                   self._log_file_path.absolute()),
                self._log_file_path.absolute())



from pathlib import Path

from exasol_integration_test_docker_environment.lib.config.build_config import build_config
from exasol_integration_test_docker_environment.lib.docker.images.create.docker_image_creator_base_task import \
    DockerImageCreatorBaseTask


class DockerLoadImageTask(DockerImageCreatorBaseTask):

    def run_task(self):
        image_archive_path = Path(build_config().cache_directory) \
            .joinpath(self.image_info.get_source_complete_name() + ".tar")
        self.logger.info("Try to load docker image %s from %s",
                         self.image_info.get_source_complete_name(), image_archive_path)
        with self._get_docker_client() as docker_client:
            with image_archive_path.open("rb") as f:
                docker_client.images.load(f)
            docker_client.images.get(self.image_info.get_source_complete_name()).tag(
                repository=self.image_info.target_repository_name,
                tag=self.image_info.get_target_complete_tag()
            )



import docker
import luigi

from exasol_integration_test_docker_environment.lib.docker import ContextDockerClient


class DockerImageTarget(luigi.Target):
    def __init__(self, image_name: str, image_tag: str):
        self.image_name = image_name
        self.image_tag = image_tag

    def get_complete_name(self):
        return f"{self.image_name}:{self.image_tag}"

    def exists(self) -> bool:
        with ContextDockerClient() as docker_client:
            try:
                image = docker_client.images.get(self.get_complete_name())
                return True
            except docker.errors.DockerException as e:
                return False



import dataclasses
from pathlib import Path
from typing import Tuple

import luigi
from docker.models.containers import Container
from exasol_bucketfs_utils_python import list_files, upload
from exasol_bucketfs_utils_python.bucket_config import BucketConfig
from exasol_bucketfs_utils_python.bucketfs_config import BucketFSConfig
from exasol_bucketfs_utils_python.bucketfs_connection_config import BucketFSConnectionConfig

# TODO add timeout, because sometimes the upload stucks
from exasol_integration_test_docker_environment.abstract_method_exception import AbstractMethodException
from exasol_integration_test_docker_environment.lib.base.docker_base_task import DockerBaseTask
from exasol_integration_test_docker_environment.lib.base.json_pickle_parameter import JsonPickleParameter
from exasol_integration_test_docker_environment.lib.base.still_running_logger import StillRunningLogger, \
    StillRunningLoggerThread
from exasol_integration_test_docker_environment.lib.data.environment_info import EnvironmentInfo
from exasol_integration_test_docker_environment.lib.test_environment.database_setup.docker_db_log_based_bucket_sync_checker import \
    DockerDBLogBasedBucketFSSyncChecker
from exasol_integration_test_docker_environment.lib.test_environment.database_setup.time_based_bucketfs_sync_waiter import \
    TimeBasedBucketFSSyncWaiter


@dataclasses.dataclass
class UploadResult:
    upload_target: str
    reused: bool


class UploadFileToBucketFS(DockerBaseTask):
    environment_name = luigi.Parameter()
    test_environment_info = JsonPickleParameter(
        EnvironmentInfo, significant=False)  # type: EnvironmentInfo
    reuse_uploaded = luigi.BoolParameter(False, significant=False)
    bucketfs_write_password = luigi.Parameter(
        significant=False, visibility=luigi.parameter.ParameterVisibility.HIDDEN)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._database_info = self.test_environment_info.database_info

    def run_task(self):
        file_to_upload = self.get_file_to_upload()
        upload_target = self.get_upload_target()
        pattern_to_wait_for = self.get_pattern_to_wait_for()
        log_file = self.get_log_file()
        sync_time_estimation = self.get_sync_time_estimation()

        with self._get_docker_client() as docker_client:
            if self._database_info.container_info is not None:
                database_container = docker_client.containers.get(
                    self._database_info.container_info.container_name)
            else:
                database_container = None
            if not self.should_be_reused(upload_target):
                self.upload_and_wait(database_container,
                                     file_to_upload,
                                     upload_target,
                                     log_file,
                                     pattern_to_wait_for,
                                     sync_time_estimation)
                self.return_object(UploadResult(
                    upload_target=upload_target,
                    reused=False
                ))
            else:
                self.logger.warning("Reusing uploaded target %s instead of file %s",
                                    upload_target, file_to_upload)
                self.write_logs("Reusing")
                self.return_object(UploadResult(
                    upload_target=upload_target,
                    reused=True
                ))

    def upload_and_wait(self, database_container,
                        file_to_upload: str, upload_target: str,
                        log_file: str, pattern_to_wait_for: str,
                        sync_time_estimation: int):
        still_running_logger = StillRunningLogger(self.logger,
                                                  "file upload of %s to %s"
                                                  % (file_to_upload, upload_target))
        thread = StillRunningLoggerThread(still_running_logger)
        thread.start()
        sync_checker = self.get_sync_checker(database_container, sync_time_estimation,
                                             log_file, pattern_to_wait_for)
        sync_checker.prepare_upload()
        try:
            output = self.upload_file(file_to_upload=file_to_upload, upload_target=upload_target)
            sync_checker.wait_for_bucketfs_sync()
            self.write_logs(output)
        finally:
            thread.stop()
            thread.join()

    def get_sync_checker(self, database_container: Container,
                         sync_time_estimation: int,
                         log_file: str,
                         pattern_to_wait_for: str):
        if database_container is not None:
            return DockerDBLogBasedBucketFSSyncChecker(
                database_container=database_container,
                log_file_to_check=log_file,
                pattern_to_wait_for=pattern_to_wait_for,
                logger=self.logger,
                bucketfs_write_password=str(self.bucketfs_write_password)
            )
        else:
            return TimeBasedBucketFSSyncWaiter(sync_time_estimation)

    def should_be_reused(self, upload_target: str):
        return self.reuse_uploaded and self.exist_file_in_bucketfs(upload_target)

    @staticmethod
    def split_upload_target(upload_target: str) -> Tuple[str, str, str]:
        upload_parts = upload_target.split("/")
        bucket_name = upload_parts[0]
        path_in_bucket = "/".join(upload_parts[1:-1])
        file_in_bucket = upload_parts[-1]
        return bucket_name, path_in_bucket, file_in_bucket

    def exist_file_in_bucketfs(self, upload_target: str) -> bool:
        self.logger.info("Check if file %s exist in bucketfs", upload_target)
        bucket_name, path_in_bucket, file_in_bucket = self.split_upload_target(upload_target)

        bucket_config = self.generate_bucket_config(bucket_name)
        try:
            files = list_files.list_files_in_bucketfs(
                bucket_config=bucket_config,
                bucket_file_path=path_in_bucket)
            return file_in_bucket in files
        except FileNotFoundError as ex:
            return False

    def generate_bucket_config(self, bucket_name: str) -> BucketConfig:
        connection_config = BucketFSConnectionConfig(
            host=self._database_info.host, port=int(self._database_info.bucketfs_port),
            user="w", pwd=str(self.bucketfs_write_password),
            is_https=False)
        bucketfs_config = BucketFSConfig(
            connection_config=connection_config,
            bucketfs_name="bfsdefault")
        bucket_config = BucketConfig(
            bucket_name=bucket_name,
            bucketfs_config=bucketfs_config)
        return bucket_config

    def upload_file(self, file_to_upload: str, upload_target: str):
        self.logger.info("upload file %s to %s",
                         file_to_upload, upload_target)
        bucket_name, path_in_bucket, file_in_bucket = self.split_upload_target(upload_target)

        bucket_config = self.generate_bucket_config(bucket_name)
        upload.upload_file_to_bucketfs(
            bucket_config=bucket_config,
            bucket_file_path=f"{path_in_bucket}/{file_in_bucket}",
            local_file_path=Path(file_to_upload))
        return f"File '{file_to_upload}' to '{upload_target}'"

    def write_logs(self, output):
        log_file = Path(self.get_log_path(), "log")
        with log_file.open("w") as file:
            file.write(output)

    def get_log_file(self) -> str:
        raise AbstractMethodException()

    def get_pattern_to_wait_for(self) -> str:
        raise AbstractMethodException()

    def get_file_to_upload(self) -> str:
        raise AbstractMethodException()

    def get_upload_target(self) -> str:
        raise AbstractMethodException()

    def get_sync_time_estimation(self) -> int:
        """Estimated time in seconds which the bucketfs needs to extract and sync a uploaded file"""
        raise AbstractMethodException()



import time

from exasol_integration_test_docker_environment.lib.base.still_running_logger import StillRunningLogger
from exasol_integration_test_docker_environment.lib.config.docker_config import source_docker_repository_config
from exasol_integration_test_docker_environment.lib.docker.images.create.docker_image_creator_base_task import \
    DockerImageCreatorBaseTask
from exasol_integration_test_docker_environment.lib.docker.images.create.utils.docker_image_target import \
    DockerImageTarget
from exasol_integration_test_docker_environment.lib.docker.images.create.utils.pull_log_handler import PullLogHandler


class DockerPullImageTask(DockerImageCreatorBaseTask):
    RETRIES = 3

    def run_task(self):
        image_target = DockerImageTarget(image_name=self.image_info.source_repository_name,
                                         image_tag=self.image_info.get_source_complete_tag())
        if source_docker_repository_config().username is not None and \
                source_docker_repository_config().password is not None:
            auth_config = {
                "username": source_docker_repository_config().username,
                "password": source_docker_repository_config().password
            }
        else:
            auth_config = None

        for i in range(self.RETRIES):
            try:
                self.pull(image_target, auth_config)
                break
            except Exception as e:
                if i < (self.RETRIES - 1):
                    self.logger.exception("Error while pulling image: %s", e)
                    time.sleep(20)
                else:
                    raise e

        with self._get_docker_client() as docker_client:
            docker_client.images.get(self.image_info.get_source_complete_name()).tag(
                repository=self.image_info.target_repository_name,
                tag=self.image_info.get_target_complete_tag()
            )

    def pull(self, image_target, auth_config):
        self.logger.info("Try to pull docker image %s", image_target.get_complete_name())
        with self._get_docker_client() as docker_client:
            output_generator = docker_client.api.pull(
                repository=image_target.image_name, tag=image_target.image_tag,
                auth_config=auth_config,
                stream=True)
            self._handle_output(output_generator, self.image_info)

    def _handle_output(self, output_generator, image_info):
        log_file_path = self.get_log_path().joinpath("pull_docker_db_image.log")
        with PullLogHandler(log_file_path, self.logger, image_info) as log_handler:
            still_running_logger = StillRunningLogger(
                self.logger, "pull image %s" % image_info.get_source_complete_name())
            for log_line in output_generator:
                still_running_logger.log()
                log_handler.handle_log_lines(log_line)



from typing import Dict

import docker
import luigi
from exasol_integration_test_docker_environment.lib.base.docker_base_task import DockerBaseTask
from exasol_integration_test_docker_environment.lib.data.docker_volume_info import DockerVolumeInfo
from exasol_integration_test_docker_environment.lib.docker.images.image_info import ImageInfo
from exasol_integration_test_docker_environment.lib.test_environment.create_certificates.analyze_certificate_container import \
    DockerCertificateContainerBuild, DockerCertificateBuildBase
from exasol_integration_test_docker_environment.lib.utils.resource_directory import ResourceDirectory
import exasol_integration_test_docker_environment.certificate_resources.container

CERTIFICATES_MOUNT_PATH = "/certificates"


class CreateSSLCertificatesTask(DockerBaseTask):
    environment_name = luigi.Parameter()
    docker_runtime = luigi.OptionalParameter(None, significant=False)
    db_container_name = luigi.Parameter(significant=False)
    network_name = luigi.Parameter()
    reuse = luigi.BoolParameter(False, significant=False)
    no_cleanup_after_success = luigi.BoolParameter(False, significant=False)
    no_cleanup_after_failure = luigi.BoolParameter(False, significant=False)
    volume_name = luigi.Parameter()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._temp_resource_directory = \
            ResourceDirectory(exasol_integration_test_docker_environment.certificate_resources.container)
        self._temp_resource_directory.create()

    def on_failure(self, exception):
        super(CreateSSLCertificatesTask, self).on_failure(exception)
        self._temp_resource_directory.cleanup()

    def on_success(self):
        super(CreateSSLCertificatesTask, self).on_success()
        self._temp_resource_directory.cleanup()

    def run_task(self):
        self.volume_info = None
        image_infos = yield from self.build_image()
        if self.reuse:
            self.logger.info("Try to reuse volume %s", self.volume_name)
            try:
                self.volume_info = self.get_volume_info(reused=True)
            except Exception as e:
                self.logger.warning(f"Tried to reuse volume {self.volume_name}, but got Exeception {e}. "
                                    "Fallback to create new volume.")
        if self.volume_info is None:
            self.volume_info = self.create_docker_volume()
            self.create_certificate(image_infos)

        self.return_object(self.volume_info)

    def build_image(self) -> Dict[str, ImageInfo]:
        task = self.create_child_task(task_class=DockerCertificateContainerBuild,
                                      certificate_container_root_directory=self._temp_resource_directory.tmp_directory)
        image_infos_future = yield from self.run_dependencies(task)
        image_infos = self.get_values_from_future(image_infos_future)
        return image_infos

    def get_volume_info(self, reused: bool) -> DockerVolumeInfo:
        with self._get_docker_client() as docker_client:
            volume_properties = docker_client.api.inspect_volume(self.volume_name)
            if volume_properties['Name'] == self.volume_name:
                return DockerVolumeInfo(volume_name=str(self.volume_name),
                                        mount_point=volume_properties['Mountpoint'], reused=reused)

    def create_docker_volume(self) -> DockerVolumeInfo:
        self.remove_volume(self.volume_name)
        with self._get_docker_client() as docker_client:
            volume = docker_client.volumes.create(
                name=self.volume_name,
            )
            volume_info = self.get_volume_info(reused=False)
        return volume_info

    def remove_volume(self, volume_name):
        try:
            with self._get_docker_client() as docker_client:
                docker_client.volumes.get(volume_name).remove()
                self.logger.info("Removed volume %s", volume_name)
        except docker.errors.NotFound:
            pass

    def cleanup_task(self, success):
        if (success and not self.no_cleanup_after_success) or \
                (not success and not self.no_cleanup_after_failure):
            try:
                self.logger.info(f"Cleaning up volume %s:", self.volume_name)
                self.remove_volume(self.volume_name)
            except Exception as e:
                self.logger.error(f"Error during removing volume %s: %s:", self.volume_name, e)

    @property
    def _construct_complete_host_name(self):
        """
        Example: 'db_container_exasol_test.db_network_exasol_test'
        """
        return f"{self.db_container_name}.{self.network_name}"

    def create_certificate(self, image_infos: Dict[str, ImageInfo]) -> None:
        certificate_container_image_info = image_infos[DockerCertificateBuildBase.GOAL]

        volumes = {
            self.volume_info.volume_name: {
                "bind": CERTIFICATES_MOUNT_PATH,
                "mode": "rw"
            }
        }

        with self._get_docker_client() as docker_client:
            try:
                test_container = \
                    docker_client.containers.create(
                        image=certificate_container_image_info.get_target_complete_name(),
                        name="certificate_resources",
                        network_mode=None,
                        command="sleep infinity",
                        detach=True,
                        volumes=volumes,
                        labels={"test_environment_name": self.environment_name,
                                "container_type": "certificate_resources"},
                        runtime=self.docker_runtime
                    )
                test_container.start()
                self.logger.info("Creating certificates...")
                cmd = f"bash /scripts/create_certificates.sh " \
                      f"{self._construct_complete_host_name} {CERTIFICATES_MOUNT_PATH}"
                exit_code, output = test_container.exec_run(cmd)
                self.logger.info(output.decode('utf-8'))
                if exit_code != 0:
                    raise RuntimeError(f"Error creating certificates:'{output.decode('utf-8')}'")
            finally:
                test_container.stop()
                test_container.remove()



import luigi
from luigi import Config

from luigi.parameter import ParameterVisibility


class ExternalDatabaseXMLRPCParameter(Config):
    external_exasol_xmlrpc_host = luigi.OptionalParameter()
    external_exasol_xmlrpc_port = luigi.IntParameter(443)
    external_exasol_xmlrpc_user = luigi.OptionalParameter()
    external_exasol_xmlrpc_cluster_name = luigi.OptionalParameter()
    external_exasol_xmlrpc_password = luigi.OptionalParameter(significant=False,
                                                              visibility=ParameterVisibility.HIDDEN)


class ExternalDatabaseHostParameter(Config):
    external_exasol_db_host = luigi.OptionalParameter()
    external_exasol_db_port = luigi.IntParameter(8563)
    external_exasol_bucketfs_port = luigi.IntParameter(6583)


class ExternalDatabaseCredentialsParameter(ExternalDatabaseHostParameter,
                                           ExternalDatabaseXMLRPCParameter):
    external_exasol_db_user = luigi.OptionalParameter()
    external_exasol_db_password = luigi.OptionalParameter(significant=False,
                                                          visibility=ParameterVisibility.HIDDEN)
    external_exasol_bucketfs_write_password = luigi.OptionalParameter(significant=False,
                                                                      visibility=ParameterVisibility.HIDDEN)



import luigi
from enum import Enum, auto
from luigi import Config

class DbOsAccess(Enum):
    """
    This enum represents different methods to access the operating system
    of the Docker Container running the Exasol database.

    The Docker Container of older database versions allowed to use
    ``docker_exec`` as well, while newer versions require SSH.

    Currently the SSH access is not implemented fully but the current test
    case will verify that when using SSH access a file with the required
    private key is generated.
    """
    DOCKER_EXEC = auto()
    SSH = auto()


class DockerDBTestEnvironmentParameter(Config):
    docker_db_image_name = luigi.OptionalParameter(None)
    docker_db_image_version = luigi.OptionalParameter(None)
    reuse_database = luigi.BoolParameter(False, significant=False)
    db_os_access = luigi.EnumParameter(DbOsAccess.DOCKER_EXEC, enum=DbOsAccess, significant=False)
    no_database_cleanup_after_success = luigi.BoolParameter(False, significant=False)
    no_database_cleanup_after_failure = luigi.BoolParameter(False, significant=False)
    database_port_forward = luigi.OptionalParameter(None, significant=False)
    bucketfs_port_forward = luigi.OptionalParameter(None, significant=False)
    mem_size = luigi.OptionalParameter("2 GiB", significant=False)
    disk_size = luigi.OptionalParameter("2 GiB", significant=False)
    nameservers = luigi.ListParameter([], significant=False)



import luigi
from luigi.parameter import ParameterVisibility

from exasol_integration_test_docker_environment.cli.options.docker_repository_options import DEFAULT_DOCKER_REPOSITORY_NAME


class source_docker_repository_config(luigi.Config):
    repository_name = luigi.Parameter(DEFAULT_DOCKER_REPOSITORY_NAME)
    tag_prefix = luigi.Parameter("")
    username = luigi.OptionalParameter(None, significant=False, visibility=ParameterVisibility.PRIVATE)
    password = luigi.OptionalParameter(None, significant=False, visibility=ParameterVisibility.PRIVATE)


class target_docker_repository_config(luigi.Config):
    repository_name = luigi.Parameter(DEFAULT_DOCKER_REPOSITORY_NAME)
    tag_prefix = luigi.Parameter("")
    username = luigi.OptionalParameter(None, significant=False, visibility=ParameterVisibility.PRIVATE)
    password = luigi.OptionalParameter(None, significant=False, visibility=ParameterVisibility.PRIVATE)


class docker_build_arguments(luigi.Config):
    transparent = luigi.DictParameter(dict())
    image_changing = luigi.DictParameter(dict())
    secret = luigi.DictParameter(dict(),
                                 description="Will not be saved somewhere, but are also assumed to be transparent",
                                 visibility=ParameterVisibility.PRIVATE)



import contextlib
import getpass
import json
import logging
import os
import warnings
from datetime import datetime
from pathlib import Path
from typing import (
    Callable,
    Tuple,
    Set,
    Optional, Any, Dict, Iterator, Iterable, List
)

import luigi
import networkx
from luigi.parameter import UnconsumedParameterWarning
from luigi.setup_logging import InterfaceLogging
from networkx import DiGraph

from exasol_integration_test_docker_environment.lib import extract_modulename_for_build_steps
from exasol_integration_test_docker_environment.lib.api.api_errors import TaskRuntimeError, TaskFailures
from exasol_integration_test_docker_environment.lib.base.dependency_logger_base_task import DependencyLoggerBaseTask
from exasol_integration_test_docker_environment.lib.base.luigi_log_config import get_luigi_log_config, get_log_path
from exasol_integration_test_docker_environment.lib.base.task_dependency import TaskDependency, DependencyState


class JobCounterSingleton(object):
    """
    We use here a Singleton to avoid a unprotected global variable.
    However, this counter needs to be global counter to guarantee unique job ids.
    This is needed in case of task that finish in less than a second, to avoid duplicated job ids
    """

    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(JobCounterSingleton, cls).__new__(cls)
            cls._instance._counter = 0
        return cls._instance

    def get_next_value(self) -> int:
        self._counter += 1
        return self._counter


def set_build_config(force_rebuild: bool,
                     force_rebuild_from: Tuple[str, ...],
                     force_pull: bool,
                     log_build_context_content: bool,
                     output_directory: str,
                     temporary_base_directory: str,
                     cache_directory: str,
                     build_name: str, ):
    luigi.configuration.get_config().set('build_config', 'force_rebuild', str(force_rebuild))
    luigi.configuration.get_config().set('build_config', 'force_rebuild_from', json.dumps(force_rebuild_from))
    luigi.configuration.get_config().set('build_config', 'force_pull', str(force_pull))
    set_output_directory(output_directory)
    if temporary_base_directory is not None:
        luigi.configuration.get_config().set('build_config', 'temporary_base_directory', temporary_base_directory)
    if cache_directory is not None:
        luigi.configuration.get_config().set('build_config', 'cache_directory', cache_directory)
    if build_name is not None:
        luigi.configuration.get_config().set('build_config', 'build_name', build_name)
    luigi.configuration.get_config().set('build_config', 'log_build_context_content', str(log_build_context_content))


def set_output_directory(output_directory):
    if output_directory is not None:
        luigi.configuration.get_config().set('build_config', 'output_directory', output_directory)


def set_docker_repository_config(docker_password: str, docker_repository_name: str, docker_username: str,
                                 tag_prefix: str,
                                 kind: str):
    config_class = f'{kind}_docker_repository_config'
    luigi.configuration.get_config().set(config_class, 'tag_prefix', tag_prefix)
    if docker_repository_name is not None:
        luigi.configuration.get_config().set(config_class, 'repository_name', docker_repository_name)
    password_environment_variable_name = f"{kind.upper()}_DOCKER_PASSWORD"
    if docker_username is not None:
        luigi.configuration.get_config().set(config_class, 'username', docker_username)
        if docker_password is not None:
            luigi.configuration.get_config().set(config_class, 'password', docker_password)
        elif password_environment_variable_name in os.environ:
            print(f"Using password from environment variable {password_environment_variable_name}")
            password = os.environ[password_environment_variable_name]
            luigi.configuration.get_config().set(config_class, 'password', password)
        else:
            password = getpass.getpass(f"{kind.capitalize()} Docker Registry Password for User %s:" % docker_username)
            luigi.configuration.get_config().set(config_class, 'password', password)


def import_build_steps(flavor_path: Tuple[str, ...]):
    # We need to load the build steps of a flavor in the commandline processor,
    # because the imported classes need to be available in all processes spawned by luigi.
    # If we  import the build steps in a Luigi Task they are only available in the worker
    # which executes this task. The build method with local scheduler of luigi uses fork
    # to create the scheduler and worker processes, such that the imported classes available
    # in the scheduler and worker processes
    import importlib.util
    for path in flavor_path:
        path_to_build_steps = Path(path).joinpath("flavor_base/build_steps.py")
        module_name_for_build_steps = extract_modulename_for_build_steps(path)
        spec = importlib.util.spec_from_file_location(module_name_for_build_steps, path_to_build_steps)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)


def generate_root_task(task_class, *args, **kwargs) -> DependencyLoggerBaseTask:
    job_counter = JobCounterSingleton().get_next_value()
    strftime = datetime.now().strftime('%Y_%m_%d_%H_%M_%S')
    params = {"job_id": f"{strftime}_{job_counter}_{task_class.__name__}"}
    params.update(kwargs)
    return task_class(**params)


def run_task(task_creator: Callable[[], DependencyLoggerBaseTask], workers: int = 2,
             task_dependencies_dot_file: Optional[str] = None,
             log_level: str = None, use_job_specific_log_file: bool = False) \
        -> Any:
    setup_worker()
    task = task_creator()
    success = False
    log_file_path = get_log_path(task.job_id)
    try:
        no_scheduling_errors = \
            _run_task_with_logging_config(
                task, log_file_path, log_level, use_job_specific_log_file, workers)
        success = not task.failed_target.exists() and no_scheduling_errors
        return _handle_task_result(no_scheduling_errors, success, task,
                                   task_dependencies_dot_file)
    except BaseException as e:
        logging.error("Going to abort the task %s" % task)
        raise e
    finally:
        if use_job_specific_log_file:
            logging.info(
                f"The detailed log of the integration-test-docker-environment can be found at: {log_file_path}")
        task.cleanup(success)


def _run_task_with_logging_config(
        task: DependencyLoggerBaseTask,
        log_file_path: Path,
        log_level: Optional[str],
        use_job_specific_log_file: bool,
        workers: int) -> bool:
    with _configure_logging(log_file_path, log_level, use_job_specific_log_file) as run_kwargs:
        no_scheduling_errors = luigi.build([task], workers=workers,
                                           local_scheduler=True, **run_kwargs)
        return no_scheduling_errors


def _handle_task_result(
        no_scheduling_errors: bool,
        success: bool,
        task: DependencyLoggerBaseTask,
        task_dependencies_dot_file: Optional[str]) -> Any:
    generate_graph_from_task_dependencies(task, task_dependencies_dot_file)
    if success:
        return task.get_result()
    elif task.failed_target.exists():
        logging.error(f"Task {task} failed. failed target exists.")
        task_failures = list(task.collect_failures().keys())
        raise TaskRuntimeError(
            msg=f"Task {task} (or any of it's subtasks) failed.",
            inner=task_failures) \
            from TaskFailures(inner=task_failures)
    elif not no_scheduling_errors:
        logging.error(f"Task {task} failed. : luigi reported a scheduling error.")
        raise TaskRuntimeError(msg=f"Task {task} failed. reason: luigi reported a scheduling error.")


@contextlib.contextmanager
def _configure_logging(
        log_file_path: Path,
        log_level: Optional[str],
        use_job_specific_log_file: bool) -> Iterator[Dict[str, str]]:
    with get_luigi_log_config(log_file_target=log_file_path,
                              log_level=log_level,
                              use_job_specific_log_file=use_job_specific_log_file) as luigi_config:
        no_configure_logging, run_kwargs = _configure_logging_parameter(
            log_level=log_level,
            luigi_config=luigi_config,
            use_job_specific_log_file=use_job_specific_log_file)
        # We need to set InterfaceLogging._configured to false,
        # because otherwise luigi doesn't accept the new config.
        InterfaceLogging._configured = False
        luigi.configuration.get_config().set('core', 'no_configure_logging', str(no_configure_logging))
        with warnings.catch_warnings():
            # This filter is necessary, because luigi uses the config no_configure_logging,
            # but doesn't define it, which seems to be a bug in luigi
            warnings.filterwarnings(action="ignore",
                                    category=UnconsumedParameterWarning,
                                    message=".*no_configure_logging.*")
            yield run_kwargs


def _configure_logging_parameter(log_level: str, luigi_config: Path, use_job_specific_log_file: bool) \
        -> Tuple[bool, Dict[str, str]]:
    if use_job_specific_log_file:
        no_configure_logging = False
        run_kwargs = {"logging_conf_file": f'{luigi_config}'}
    else:
        no_configure_logging = True
        run_kwargs = {}
    return no_configure_logging, run_kwargs


def generate_graph_from_task_dependencies(task: DependencyLoggerBaseTask, task_dependencies_dot_file: Optional[str]):
    if task_dependencies_dot_file is not None:
        print(f"Generate Task Dependency Graph to {task_dependencies_dot_file}")
        print()
        dependencies = collect_dependencies(task)
        g = DiGraph()
        for dependency in dependencies:
            g.add_node(dependency.source.formatted_id, label=dependency.source.formatted_representation)
            g.add_node(dependency.target.formatted_id, label=dependency.target.formatted_representation)
            g.add_edge(dependency.source.formatted_id, dependency.target.formatted_id,
                       dependency=dependency.formatted,
                       label=f"\"type={dependency.type}, index={dependency.index}\"")
        networkx.nx_pydot.write_dot(g, task_dependencies_dot_file)


def collect_dependencies(task: DependencyLoggerBaseTask) -> Set[TaskDependency]:
    dependencies = set()
    for root, directories, files in os.walk(task._get_dependencies_path_for_job()):
        for file in files:
            file_path = Path(root).joinpath(file)
            with open(file_path) as f:
                for line in f.readlines():
                    task_dependency = TaskDependency.from_json(line)  # type: TaskDependency
                    if task_dependency.state == DependencyState.requested.name:
                        dependencies.add(task_dependency)
    return dependencies


def setup_worker():
    luigi.configuration.get_config().set('worker', 'wait_interval', str(0.1))
    luigi.configuration.get_config().set('worker', 'wait_jitter', str(0.5))


def add_options(options):
    def _add_options(func):
        for option in reversed(options):
            func = option(func)
        return func

    return _add_options


def cli_function(func):
    """Decorator: Register a function as having a cli equivalent"""
    func.__cli_function__ = True
    return func


def no_cli_function(func):
    """Decorator: Register a function as noy having a cli equivalent"""
    func.__cli_function__ = False
    return func



import subprocess
from typing import Optional

from exasol_integration_test_docker_environment.cli.options.test_environment_options import LATEST_DB_VERSION
from exasol_integration_test_docker_environment.lib.data.environment_info import EnvironmentInfo
from exasol_integration_test_docker_environment.testing.utils import check_db_version_from_env


class ExaslctDockerTestEnvironment:
    def __init__(self, name: str, database_host: str,
                 db_username: str, db_password: str,
                 bucketfs_username: str, bucketfs_password: str,
                 database_port: int, bucketfs_port: int,
                 environment_info: Optional[EnvironmentInfo] = None,
                 completed_process: Optional[subprocess.CompletedProcess] = None):
        self.db_password = db_password
        self.db_username = db_username
        self.database_port = database_port
        self.bucketfs_port = bucketfs_port
        self.bucketfs_username = bucketfs_username
        self.bucketfs_password = bucketfs_password
        self.database_host = database_host
        self.name = name
        self.environment_info = environment_info
        self.completed_process = completed_process
        self.clean_up = None
        self.docker_db_image_version = check_db_version_from_env() or LATEST_DB_VERSION

    def close(self):
        if self.clean_up is not None:
            self.clean_up()



import contextlib
import os
import tempfile
from pathlib import Path
from typing import Optional, Callable

import jinja2
import logging
from exasol_integration_test_docker_environment.lib import PACKAGE_NAME
from exasol_integration_test_docker_environment.lib.config.build_config import build_config

LOG_ENV_VARIABLE_NAME = "EXA_BUILD_LOG"

PROPAGATE = "PROPAGATE"

FILTERS = "FILTERS"

HANDLERS = "HANDLERS"

LOG_LEVEL = "LOG_LEVEL"

LUIGI_INTERFACE_LOGGER = "luigi-interface"

LUIGI_LOGGER = "luigi"


def get_log_path(job_id: str) -> Path:
    """
    Retrieve the log-file path. Default path is $output_path/jobs/logs/main.log, but can be overwritten by
    the environment variable LOG_ENV_VARIABLE_NAME.
    """
    main_log_path = Path(build_config().output_directory) / "jobs" / job_id / "logs"
    main_log_path.mkdir(parents=True, exist_ok=True)
    def_log_path = main_log_path / "main.log"
    env_log_path = os.getenv(LOG_ENV_VARIABLE_NAME)
    if env_log_path is not None:
        log_path = Path(env_log_path)
    else:
        log_path = def_log_path
    return log_path


@contextlib.contextmanager
def restore_logger(logger_creator: Callable[[], logging.Logger]):
    before_logger = logger_creator()
    logger_info = {
        LOG_LEVEL: before_logger.level,
        HANDLERS: list(before_logger.handlers),
        FILTERS: list(before_logger.filters),
        PROPAGATE: before_logger.propagate
    }
    yield
    after_logger = logger_creator()
    after_logger.level = logger_info[LOG_LEVEL]
    after_logger.handlers = logger_info[HANDLERS]
    after_logger.filters = logger_info[FILTERS]
    after_logger.propagate = logger_info[PROPAGATE]


@contextlib.contextmanager
def get_luigi_log_config(log_file_target: Path,
                         use_job_specific_log_file: bool,
                         log_level: Optional[str] = None) -> Path:
    """
    Yields a context manager containing the path of the log-config file.
    log_file_target contains the location of the log-file.
    console_log_level indicates the log_level (@see https://docs.python.org/3/library/logging.html#logging-levels)
    The log-level for the log-file is always logging.DEBUG!
    """
    if log_level is None and use_job_specific_log_file:
        log_level = logging.getLevelName(logging.WARNING)
    env = jinja2.Environment(loader=jinja2.PackageLoader(PACKAGE_NAME),
                             autoescape=jinja2.select_autoescape())
    template = env.get_template("luigi_log.conf")
    rendered_template = template.render(console_log_level=log_level,
                                        log_file_target=str(log_file_target))
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_luigi_conf_path = Path(temp_dir) / "luigi_log.conf"
        with open(temp_luigi_conf_path, "w") as f:
            f.write(rendered_template)
        with restore_logger(logger_creator=lambda: logging.root), \
                restore_logger(logger_creator=lambda: logging.getLogger(LUIGI_INTERFACE_LOGGER)), \
                restore_logger(logger_creator=lambda: logging.getLogger(LUIGI_LOGGER)):
            if log_level is not None and not use_job_specific_log_file:
                logging.getLogger(LUIGI_INTERFACE_LOGGER).level = logging.getLevelName(log_level)
                logging.getLogger(LUIGI_LOGGER).level = logging.getLevelName(log_level)

            yield temp_luigi_conf_path



from collections import ChainMap
from dataclasses import dataclass
from typing import Generic, List, Optional, TypeVar

from pyexasol.connection import ExaConnection

T = TypeVar("T")


@dataclass(frozen=True)
class Option(Generic[T]):
    name: str
    prefix: str
    type: T
    default: Optional[T] = None
    help_text: str = ""

    @property
    def env(self):
        """Environment variable name"""

        def normalize(name):
            name = name.replace("-", "_")
            name = name.upper()
            return name

        return f"{normalize(self.prefix)}_{normalize(self.name)}"

    @property
    def cli(self):
        """Cli argument name"""

        def normalize(name):
            name = name.replace("_", "-")
            name = name.lower()
            return name

        return f"--{normalize(self.prefix)}-{normalize(self.name)}"

    @property
    def pytest(self):
        """Pytest option name"""

        def normalize(name):
            name = name.replace("-", "_")
            name = name.lower()
            return name

        return f"{normalize(self.prefix)}_{normalize(self.name)}"

    @property
    def help(self):
        """Help text including information about default value."""
        if not self.default:
            return f"{self.help_text}."
        return f"{self.help_text} (default: {self.default})."


class OptionGroup:
    """
    Wraps a set of pytest options.
    """

    def __init__(self, prefix, options):
        self._prefix = prefix
        self._options = tuple(Option(prefix=prefix, **kwargs) for kwargs in options)
        self._default = {o.name: o.default for o in self._options}
        self._env = {}
        self._cli = {}
        self._kwargs = ChainMap(self._cli, self._env, self._default)

    @property
    def prefix(self):
        """The option group prefix."""
        return self._prefix

    @property
    def options(self):
        """A tuple of all options which are part of this group."""
        return self._options

    def kwargs(self, environment, cli_arguments):
        """
        Given the default values, the passed environment and cli arguments it will
        take care of the prioritization for the option values in regard of their
        source(s) and return a kwargs dictionary with all options and their
        appropriate value.
        """
        env = {
            o.name: o.type(environment[o.env])
            for o in self._options
            if o.env in environment
        }
        cli = {
            o.name: getattr(cli_arguments, o.pytest)
            for o in self.options
            if hasattr(cli_arguments, o.pytest)
            and getattr(cli_arguments, o.pytest) is not None
        }
        self._env.update(env)
        self._cli.update(cli)
        return self._kwargs


@dataclass
class Exasol:
    """Exasol database configuration"""

    host: str
    port: int
    username: str
    password: str


@dataclass
class BucketFs:
    """Bucketfs configuration"""

    url: str
    username: str
    password: str


@dataclass
class Itde:
    """Itde configuration settings"""

    db_version: str
    schemas: List[str]


@dataclass
class TestConfig:
    """Full test configuration for itde based tests"""

    db: Exasol
    bucketfs: BucketFs
    itde: Itde
    ctrl_connection: ExaConnection



from typing import Optional

from exasol_integration_test_docker_environment.lib.base.info import Info
from exasol_integration_test_docker_environment.lib.data.container_info import ContainerInfo
from exasol_integration_test_docker_environment.lib.data.database_info import DatabaseInfo
from exasol_integration_test_docker_environment.lib.data.docker_network_info import DockerNetworkInfo


class EnvironmentInfo(Info):

    def __init__(self,
                 name: str, env_type: str,
                 database_info: DatabaseInfo,
                 test_container_info: Optional[ContainerInfo],
                 network_info: DockerNetworkInfo):
        self.name = name
        self.type = env_type
        self.test_container_info = test_container_info
        self.database_info = database_info
        self.network_info = network_info



import luigi

from exasol_integration_test_docker_environment.lib.base.docker_base_task import DockerBaseTask
from exasol_integration_test_docker_environment.lib.base.json_pickle_parameter import JsonPickleParameter
from exasol_integration_test_docker_environment.lib.data.database_credentials import DatabaseCredentialsParameter
from exasol_integration_test_docker_environment.lib.data.database_info import DatabaseInfo


class WaitForTestExternalDatabase(DockerBaseTask,
                                  DatabaseCredentialsParameter):
    environment_name = luigi.Parameter()
    database_info = JsonPickleParameter(DatabaseInfo, significant=False)  # type: DatabaseInfo
    db_startup_timeout_in_seconds = luigi.IntParameter(1 * 60, significant=False)
    attempt = luigi.IntParameter(1)

    def run_task(self):
        # Since we can't assume that the test container exists, we cannot connect easily here
        # to the external database (correct way would be by using an SQL client).
        # For now, we simply assume that the external database is already ready and return just True.
        self.return_object(True)



import logging
from pathlib import Path, PurePath

import luigi

from exasol_integration_test_docker_environment.abstract_method_exception import AbstractMethodException
from exasol_integration_test_docker_environment.lib.base.docker_base_task import DockerBaseTask
from exasol_integration_test_docker_environment.lib.base.json_pickle_parameter import JsonPickleParameter
from exasol_integration_test_docker_environment.lib.data.database_credentials import DatabaseCredentialsParameter
from exasol_integration_test_docker_environment.lib.data.environment_info import EnvironmentInfo


class PopulateTestDataToDatabase(DockerBaseTask, DatabaseCredentialsParameter):
    logger = logging.getLogger('luigi-interface')

    environment_name = luigi.Parameter()
    test_environment_info = JsonPickleParameter(
        EnvironmentInfo, significant=False)  # type: EnvironmentInfo

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._test_container_info = self.test_environment_info.test_container_info
        self._database_info = self.test_environment_info.database_info

    def run_task(self):
        self.logger.warning("Uploading data")
        username = self.db_user
        password = self.db_password
        data_path_within_test_container = self.get_data_path_within_test_container()
        data_file_within_data_path = self.get_data_file_within_data_path()
        with self._get_docker_client() as docker_client:
            test_container = docker_client.containers.get(self._test_container_info.container_name)
            cmd = f"cd {data_path_within_test_container}; " \
                  f"$EXAPLUS -c '{self._database_info.host}:{self._database_info.db_port}' " \
                  f"-x -u '{username}' -p '{password}' -f {data_file_within_data_path} " \
                  f"-jdbcparam 'validateservercertificate=0'"

            bash_cmd = f"""bash -c "{cmd}" """
            exit_code, output = test_container.exec_run(cmd=bash_cmd)
        self.write_logs(output.decode("utf-8"))
        if exit_code != 0:
            raise Exception("Failed to populate the database with data.\nLog: %s" % cmd + "\n" + output.decode("utf-8"))

    def write_logs(self, output: str):
        log_file = Path(self.get_log_path(), "log")
        with log_file.open("w") as file:
            file.write(output)

    def get_data_path_within_test_container(self) -> PurePath:
        raise AbstractMethodException()

    def get_data_file_within_data_path(self) -> PurePath:
        raise AbstractMethodException()



import luigi

from exasol_integration_test_docker_environment.lib.base.docker_base_task import DockerBaseTask
from exasol_integration_test_docker_environment.lib.base.json_pickle_parameter import JsonPickleParameter
from exasol_integration_test_docker_environment.lib.docker.images.image_info import ImageInfo


class DockerImageCreatorBaseTask(DockerBaseTask):
    image_name = luigi.Parameter()
    # ParameterVisibility needs to be hidden instead of private, because otherwise a MissingParameter gets thrown
    image_info = JsonPickleParameter(ImageInfo,
                                     visibility=luigi.parameter.ParameterVisibility.HIDDEN,
                                     significant=True)  # type:ImageInfo



import copy
import importlib

import luigi

from exasol_integration_test_docker_environment.lib.base.docker_base_task import DockerBaseTask
from exasol_integration_test_docker_environment.lib.base.json_pickle_parameter import JsonPickleParameter
from exasol_integration_test_docker_environment.lib.docker.images.create.docker_image_build_task import \
    DockerBuildImageTask
from exasol_integration_test_docker_environment.lib.docker.images.create.docker_image_load_task import \
    DockerLoadImageTask
from exasol_integration_test_docker_environment.lib.docker.images.create.docker_image_pull_task import \
    DockerPullImageTask
from exasol_integration_test_docker_environment.lib.docker.images.image_info import ImageInfo, ImageState
from exasol_integration_test_docker_environment.lib.docker.images.required_task_info import RequiredTaskInfoDict, \
    RequiredTaskInfo


class DockerCreateImageTask(DockerBaseTask):
    image_name = luigi.Parameter()
    # ParameterVisibility needs to be hidden instead of private, because otherwise a MissingParameter gets thrown
    image_info = JsonPickleParameter(ImageInfo,
                                     visibility=luigi.parameter.ParameterVisibility.HIDDEN,
                                     significant=True)  # type: ImageInfo

    def run_task(self):
        new_image_info = yield from self.build(self.image_info)
        self.return_object(new_image_info)

    def build(self, image_info: ImageInfo):
        if image_info.image_state == ImageState.NEEDS_TO_BE_BUILD.name:
            task = self.create_child_task(DockerBuildImageTask,
                                          image_name=self.image_name,
                                          image_info=image_info)
            yield from self.run_dependencies(task)
            image_info.image_state = ImageState.WAS_BUILD.name  # TODO clone and change
            return image_info
        elif image_info.image_state == ImageState.CAN_BE_LOADED.name:
            task = self.create_child_task(DockerLoadImageTask,
                                          image_name=self.image_name,
                                          image_info=image_info)
            yield from self.run_dependencies(task)
            image_info.image_state = ImageState.WAS_LOADED.name
            return image_info
        elif image_info.image_state == ImageState.REMOTE_AVAILABLE.name:
            task = self.create_child_task(DockerPullImageTask,
                                          image_name=self.image_name,
                                          image_info=image_info)
            yield from self.run_dependencies(task)
            image_info.image_state = ImageState.WAS_PULLED.name
            return image_info
        elif image_info.image_state == ImageState.TARGET_LOCALLY_AVAILABLE.name:
            image_info.image_state = ImageState.USED_LOCAL.name
            return image_info
        elif image_info.image_state == ImageState.SOURCE_LOCALLY_AVAILABLE.name:
            image_info.image_state = ImageState.WAS_TAGED.name
            self.rename_source_image_to_target_image(image_info)
            return image_info
        else:
            raise Exception("Task %s: Image state %s not supported for image %s",
                            self.task_id, image_info.image_state, image_info.get_target_complete_name())

    def rename_source_image_to_target_image(self, image_info):
        with self._get_docker_client() as docker_client:
            docker_client.images.get(image_info.get_source_complete_name()).tag(
                repository=image_info.target_repository_name,
                tag=image_info.get_target_complete_tag()
            )


class DockerCreateImageTaskWithDeps(DockerCreateImageTask):
    # ParameterVisibility needs to be hidden instead of private, because otherwise a MissingParameter gets thrown
    required_task_infos = JsonPickleParameter(RequiredTaskInfoDict,
                                              visibility=luigi.parameter.ParameterVisibility.HIDDEN,
                                              significant=True)  # type: RequiredTaskInfoDict

    def register_required(self):
        self.required_tasks = {key: self.create_required_task(required_task_info)
                               for key, required_task_info
                               in self.required_task_infos.infos.items()}
        self.futures = self.register_dependencies(self.required_tasks)

    def create_required_task(self, required_task_info: RequiredTaskInfo) -> DockerCreateImageTask:
        module = importlib.import_module(required_task_info.module_name)
        class_ = getattr(module, required_task_info.class_name)
        instance = self.create_child_task(class_, **required_task_info.params)
        return instance

    def run_task(self):
        image_infos = self.get_values_from_futures(self.futures)
        image_info = copy.copy(self.image_info)
        image_info.depends_on_images = image_infos
        new_image_info = yield from self.build(image_info)
        self.return_object(new_image_info)



from datetime import datetime
from enum import Enum
from typing import Dict, Any

from exasol_integration_test_docker_environment.lib.base.info import Info


class ImageState(Enum):
    NOT_EXISTING = 0,
    # After analyze phase or if build phase did touch the image
    NEEDS_TO_BE_BUILD = 1,
    TARGET_LOCALLY_AVAILABLE = 2,
    SOURCE_LOCALLY_AVAILABLE = 3,
    REMOTE_AVAILABLE = 4,
    CAN_BE_LOADED = 5,
    # After build phase
    WAS_BUILD = 6,
    USED_LOCAL = 7,
    WAS_PULLED = 8,
    WAS_LOADED = 9
    WAS_TAGED = 10


class ImageDescription:
    def __init__(self,
                 dockerfile: str,
                 image_changing_build_arguments: Dict[str, Any],
                 transparent_build_arguments: Dict[str, Any],
                 mapping_of_build_files_and_directories: Dict[str, str]):
        self.transparent_build_arguments = transparent_build_arguments
        self.image_changing_build_arguments = image_changing_build_arguments
        self.mapping_of_build_files_and_directories = mapping_of_build_files_and_directories
        self.dockerfile = dockerfile

    def __repr__(self):
        return f"{self.__class__.__name__}({self.__dict__})"


class ImageInfo(Info):
    DOCKER_TAG_LENGTH_LIMIT = 128
    MAX_TAG_SURPLUS = 30

    def __init__(self,
                 source_repository_name: str, target_repository_name: str,
                 source_tag: str, target_tag: str,
                 hash_value: str, commit: str,
                 image_description: ImageDescription,
                 build_name: str = "",
                 build_date_time: datetime = datetime.utcnow(),
                 image_state: ImageState = ImageState.NOT_EXISTING,
                 depends_on_images: Dict[str, "ImageInfo"] = None):
        self.build_name = build_name
        self.date_time = str(build_date_time)
        self.commit = commit
        self.target_repository_name = target_repository_name
        self.source_repository_name = source_repository_name
        self.image_description = image_description
        if isinstance(image_state, ImageState):
            self.image_state = image_state.name
        elif isinstance(image_state, str):
            self.image_state = ImageState[image_state].name
        elif image_state is None:
            self.image_state = None
        else:
            raise TypeError(f"{type(image_state)} for image_state not supported")
        self.depends_on_images = depends_on_images
        self.source_tag = source_tag
        self.target_tag = target_tag
        self.hash = hash_value
        self.check_complete_tag_length(self.source_tag)
        self.check_complete_tag_length(self.target_tag)

    def check_complete_tag_length(self, tag):
        complete_tag_length_limit = self.DOCKER_TAG_LENGTH_LIMIT + self.MAX_TAG_SURPLUS
        complete_tag = self._create_complete_tag(tag)
        if len(complete_tag) > complete_tag_length_limit:
            raise Exception(f"Complete Tag to long by {len(complete_tag) - complete_tag_length_limit}:  {complete_tag}")

    def get_target_complete_name(self):
        return f"{self.target_repository_name}:{self.get_target_complete_tag()}"

    def get_source_complete_name(self):
        return f"{self.source_repository_name}:{self.get_source_complete_tag()}"

    def get_source_complete_tag(self):
        return self._create_truncated_complete_tag(self.source_tag)

    def get_target_complete_tag(self):
        return self._create_truncated_complete_tag(self.target_tag)

    def _create_truncated_complete_tag(self, tag: str) -> str:
        # we must truncate the tag to 128 characters, because this is the limit of docker tags
        # refer here https://docs.docker.com/engine/reference/commandline/tag/
        complete_tag = self._create_complete_tag(tag)
        truncated_tag = complete_tag[:128]
        return truncated_tag

    def _create_complete_tag(self, tag):
        if self.hash == "":
            return f"{tag}"
        else:
            return f"{tag}_{self.hash}"



import json
import multiprocessing as mp


from exasol_integration_test_docker_environment.lib.docker import ContextDockerClient
from exasol_integration_test_docker_environment.lib.logging.abstract_log_handler import AbstractLogHandler


class DockerRegistryImageCheckerPullLogHandler(AbstractLogHandler):

    def __init__(self):
        super().__init__(None, None)
        self.result = False

    def handle_log_line(self, log_line, error: bool = False):
        json_output = json.loads(log_line)
        # TODO logging
        if "status" in json_output and json_output["status"].startswith("Pulling"):
            return True
        elif "errorDetail" in json_output:
            return False


class DockerRegistryImageChecker:
    def map(self, image: str, queue: mp.Queue):

        with ContextDockerClient() as docker_client:
            try:
                generator = docker_client.api.pull(repository=image, stream=True)
                for log_line in generator:
                    queue.put(log_line)
                queue.put(None)
            except Exception as e:
                queue.put(e)

    def check(self, image: str):
        log_handler = DockerRegistryImageCheckerPullLogHandler()
        queue = mp.Queue()
        process = mp.Process(target=self.map, args=(image, queue))
        process.start()
        try:
            while True:
                value = queue.get()
                if isinstance(value, Exception):
                    raise value
                elif isinstance(value, bytes):
                    return any(log_handler.handle_log_lines(value))
                elif value is None:
                    return False
                else:
                    raise RuntimeError(f"Should not happen. Programming Error. Unknown value {value}")
        finally:
            process.terminate()



import docker

from exasol_integration_test_docker_environment.lib.config.docker_config import source_docker_repository_config, \
    target_docker_repository_config


class ContextDockerClient(object):
    def __init__(self, **kwargs):
        self.kwargs = kwargs
        self._client = None

    def __enter__(self):

        self._client = docker.from_env(**self.kwargs)
        self.login(self._client, source_docker_repository_config())
        self.login(self._client, target_docker_repository_config())
        return self._client

    def login(self, client: docker.client.DockerClient, docker_repository_config):
        if docker_repository_config.username is not None and \
                docker_repository_config.password is not None:
            client.login(username=docker_repository_config.username, password=docker_repository_config.password)

    def __exit__(self, type_, value, traceback):
        if self._client is not None:
            self._client.close()
        self._client = None



from collections.abc import Mapping


class FrozenDictToDict():
    def convert(self, obj):
        if isinstance(obj, Mapping):
            return {k: self.convert(v) for k, v in obj.items()}
        elif isinstance(obj, (list, tuple)):
            return [self.convert(o) for o in obj]
        else:
            return obj



import os
from typing import Tuple

import pyexasol
import pytest

from pytest_itde import config

EXASOL = config.OptionGroup(
    prefix="exasol",
    options=(
        {
            "name": "host",
            "type": str,
            "default": "localhost",
            "help_text": "Host to connect to",
        },
        {
            "name": "port",
            "type": int,
            "default": 8888,
            "help_text": "Port on which the exasol db is listening",
        },
        {
            "name": "username",
            "type": str,
            "default": "SYS",
            "help_text": "Username used to authenticate against the exasol db",
        },
        {
            "name": "password",
            "type": str,
            "default": "exasol",
            "help_text": "Password used to authenticate against the exasol db",
        },
    ),
)

BUCKETFS = config.OptionGroup(
    prefix="bucketfs",
    options=(
        {
            "name": "url",
            "type": str,
            "default": "http://127.0.0.1:6666",
            "help_text": "Base url used to connect to the bucketfs service",
        },
        {
            "name": "username",
            "type": str,
            "default": "w",
            "help_text": "Username used to authenticate against the bucketfs service",
        },
        {
            "name": "password",
            "type": str,
            "default": "write",
            "help_text": "Password used to authenticate against the bucketfs service",
        },
    ),
)


def TestSchemas(value) -> Tuple[str]:
    """
    Parses a test schema spec string.

    Args:
        value: spec string for the test schemas.
        e.g. "Schema1, Schema2, " or "Schema"
    """

    seperator = ","
    if seperator in value:
        schemas = value.split(seperator)
    else:
        schemas = [value]
    schemas = (s.strip() for s in schemas)
    schemas = (s for s in schemas if s != "")
    return tuple(schemas)


ITDE = config.OptionGroup(
    prefix="itde",
    options=(
        {
            "name": "db_version",
            "type": str,
            "default": "8.18.1",
            "help_text": "DB version to start, if value is 'external' an existing instance will be used",
        },
        {
            "name": "schemas",
            "type": TestSchemas,
            "default": ("TEST", "TEST_SCHEMA"),
            "help_text": "Schemas which should be created for the session",
        },
    ),
)


@pytest.fixture(scope="session")
def exasol_config(request) -> config.Exasol:
    """Returns the configuration settings of the exasol db for this session."""
    cli_arguments = request.config.option
    kwargs = EXASOL.kwargs(os.environ, cli_arguments)
    return config.Exasol(**kwargs)


@pytest.fixture(scope="session")
def bucketfs_config(request) -> config.BucketFs:
    """Returns the configuration settings of the bucketfs for this session."""
    cli_arguments = request.config.option
    kwargs = BUCKETFS.kwargs(os.environ, cli_arguments)
    return config.BucketFs(**kwargs)


@pytest.fixture(scope="session")
def itde_config(request) -> config.Itde:
    """Returns the configuration settings of the ITDE for this session."""
    cli_arguments = request.config.option
    kwargs = ITDE.kwargs(os.environ, cli_arguments)
    return config.Itde(**kwargs)


@pytest.fixture(scope="session")
def connection_factory():
    """
    Returns a database connection factory.

    Attention:
        All created connections will be cleaned up (closed) at the end of the session.
    """
    connections = []

    def factory(config: config.Exasol):
        con = pyexasol.connect(
            dsn=f"{config.host}:{config.port}",
            user=config.username,
            password=config.password,
        )
        connections.append(con)
        return con

    yield factory

    for connection in connections:
        connection.close()


@pytest.fixture(scope="session")
def _bootstrap_db(itde_config, exasol_config, bucketfs_config):
    """Bootstraps the database should not be used from outside the itde plugin."""

    def nop():
        pass

    def start_db(name, itde, exasol, bucketfs):
        from urllib.parse import urlparse

        from exasol_integration_test_docker_environment.lib import api

        bucketfs_url = urlparse(bucketfs.url)
        _, cleanup_function = api.spawn_test_environment(
            environment_name=name,
            database_port_forward=exasol.port,
            bucketfs_port_forward=bucketfs_url.port,
            db_mem_size="4GB",
            docker_db_image_version=itde.db_version,
        )
        return cleanup_function

    db_name = "pytest_exasol_db"
    bootstrap_db = itde_config.db_version != "external"

    start = (
        lambda: start_db(db_name, itde_config, exasol_config, bucketfs_config)
        if bootstrap_db
        else lambda: nop
    )
    stop = start()
    yield
    stop()


@pytest.fixture(scope="session")
def itde(
    _bootstrap_db,
    itde_config,
    exasol_config,
    bucketfs_config,
    connection_factory,
) -> config.TestConfig:
    """Starts a docker based test environment and returns the associated test config."""
    connection = connection_factory(exasol_config)

    for schema in itde_config.schemas:
        connection.execute(f"DROP SCHEMA IF EXISTS {schema} CASCADE;")
        connection.execute(f"CREATE SCHEMA {schema};")
        connection.commit()

    yield config.TestConfig(
        db=exasol_config,
        bucketfs=bucketfs_config,
        itde=itde_config,
        ctrl_connection=connection,
    )

    for schema in itde_config.schemas:
        connection.execute(f"DROP SCHEMA IF EXISTS {schema} CASCADE;")
        connection.commit()


OPTION_GROUPS = (EXASOL, BUCKETFS, ITDE)


def _add_option_group(parser, group):
    parser_group = parser.getgroup(group.prefix)
    for option in group.options:
        parser_group.addoption(
            option.cli,
            type=option.type,
            help=option.help,
        )


def pytest_addoption(parser):
    for group in OPTION_GROUPS:
        _add_option_group(parser, group)



from exasol_integration_test_docker_environment.lib.base.dependency_logger_base_task import DependencyLoggerBaseTask
from exasol_integration_test_docker_environment.lib.base.docker_parameter import DockerParameter
from exasol_integration_test_docker_environment.lib.docker import ContextDockerClient


class DockerBaseTask(DependencyLoggerBaseTask, DockerParameter):

    def _get_docker_client(self):
        return ContextDockerClient(timeout=self.timeout)



import os
import pathlib
import shutil
import textwrap
from pathlib import Path

from jinja2 import Template

from exasol_integration_test_docker_environment.lib.config.build_config import build_config
from exasol_integration_test_docker_environment.lib.docker.images.image_info import ImageInfo, ImageState


class BuildContextCreator:
    def __init__(self,
                 temp_directory,
                 image_info: ImageInfo,
                 log_file_path: Path):
        self._image_info = image_info
        self._log_file_path = log_file_path
        self._image_info_of_dependencies = self._image_info.depends_on_images
        self._image_description = self._image_info.image_description
        self._temp_directory = temp_directory

    def prepare_build_context_to_temp_dir(self):
        self._copy_build_files_and_directories()
        self._prepare_dockerfile()
        self._prepare_image_info()
        self._log_build_context()

    def _prepare_image_info(self):
        self._image_info.image_state = ImageState.WAS_BUILD.name
        with open(self._temp_directory + "/image_info", "wt") as file:
            file.write(self._image_info.to_json())

    def _prepare_dockerfile(self):
        with open(self._image_description.dockerfile, "rt") as file:
            dockerfile_content = file.read()
        template = Template(dockerfile_content)
        image_names_of_dependencies = \
            {key: image_info.get_target_complete_name()
             for key, image_info
             in self._image_info_of_dependencies.items()}
        final_dockerfile = template.render(**image_names_of_dependencies)
        final_dockerfile += textwrap.dedent(f"""
        RUN mkdir -p /build_info/image_info
        COPY image_info /build_info/image_info/{self._image_info.target_tag}
        RUN mkdir -p /build_info/dockerfiles
        COPY Dockerfile /build_info/dockerfiles/{self._image_info.target_tag}
        """)
        with open(self._temp_directory + "/Dockerfile", "wt") as file:
            file.write(final_dockerfile)

    def _copy_build_files_and_directories(self):
        for dest, src in self._image_description.mapping_of_build_files_and_directories.items():
            src_path = pathlib.Path(src)
            dest_path = self._temp_directory + "/" + dest
            if src_path.is_dir():
                shutil.copytree(src, dest_path)
            elif src_path.is_file():
                shutil.copy2(src, dest_path)
            else:
                raise Exception("Source path %s is neither a file or a directory" % src)

    def _log_build_context(self):
        if build_config().log_build_context_content:
            build_context_log_file = self._log_file_path.joinpath("docker-build-context.log")
            with build_context_log_file.open("wt") as log_file:
                for file in self._get_files_in_build_context(self._temp_directory):
                    log_file.write(file)
                    log_file.write("\n")
            dockerfile_log_file = self._log_file_path.joinpath("Dockerfile.generated")
            shutil.copy(self._temp_directory + "/Dockerfile", str(dockerfile_log_file))

    def _get_files_in_build_context(self, temp_directory):
        return [os.path.join(r, file) for r, d, f in os.walk(temp_directory) for file in f]



from typing import Set, Dict

import luigi

from exasol_integration_test_docker_environment.lib.config.docker_config import target_docker_repository_config, \
    source_docker_repository_config
from exasol_integration_test_docker_environment.lib.docker.images.create.docker_build_base import DockerBuildBase
from exasol_integration_test_docker_environment.lib.docker.images.create.docker_image_analyze_task import \
    DockerAnalyzeImageTask
from exasol_integration_test_docker_environment.lib.docker.images.push.docker_push_parameter import DockerPushParameter
from exasol_integration_test_docker_environment.lib.docker.images.push.push_task_creator_for_build_tasks import \
    PushTaskCreatorFromBuildTasks

TAG_SUFFIX = "certificate_resources"


class AnalyzeCertificateContainer(DockerAnalyzeImageTask):
    certificate_container_root_directory = luigi.Parameter()

    def get_target_repository_name(self) -> str:
        return f"""{target_docker_repository_config().repository_name}"""

    def get_source_repository_name(self) -> str:
        return f"""{source_docker_repository_config().repository_name}"""

    def get_source_image_tag(self):
        if source_docker_repository_config().tag_prefix != "":
            return f"{source_docker_repository_config().tag_prefix}_{TAG_SUFFIX}"
        else:
            return TAG_SUFFIX

    def get_target_image_tag(self):
        if target_docker_repository_config().tag_prefix != "":
            return f"{target_docker_repository_config().tag_prefix}_{TAG_SUFFIX}"
        else:
            return TAG_SUFFIX

    def get_mapping_of_build_files_and_directories(self):
        return {"create_certificates.sh": f"{self.certificate_container_root_directory}/create_certificates.sh"}

    def get_dockerfile(self):
        return f"{self.certificate_container_root_directory}/Dockerfile"

    def is_rebuild_requested(self) -> bool:
        return False


class DockerCertificateBuildBase(DockerBuildBase):
    GOAL = "certificate-container"
    certificate_container_root_directory = luigi.Parameter()

    def get_goal_class_map(self) -> Dict[str, DockerAnalyzeImageTask]:
        goal_class_map = {self.GOAL: self.create_child_task(task_class=AnalyzeCertificateContainer,
                                                            certificate_container_root_directory=
                                                            self.certificate_container_root_directory)}
        return goal_class_map

    def get_default_goals(self) -> Set[str]:
        goals = {self.GOAL}
        return goals

    def get_goals(self):
        goals = {self.GOAL}
        return goals


class DockerCertificateContainerBuild(DockerCertificateBuildBase):

    def run_task(self):
        build_tasks = self.create_build_tasks(False)
        image_infos_futures = yield from self.run_dependencies(build_tasks)
        image_infos = self.get_values_from_futures(image_infos_futures)
        self.return_object(image_infos)


class DockerTestContainerPush(DockerCertificateBuildBase, DockerPushParameter):

    def run_task(self):
        build_tasks = self.create_build_tasks(shortcut_build=not self.push_all)
        push_task_creator = PushTaskCreatorFromBuildTasks(self)
        push_tasks = push_task_creator.create_tasks_for_build_tasks(build_tasks)
        image_info_future = yield from self.run_dependencies(push_tasks)
        image_info = self.get_values_from_futures(image_info_future)
        self.return_object(image_info)



import logging
from typing import List

from exasol_integration_test_docker_environment.lib.docker import ContextDockerClient


def remove_docker_container(containers: List[str]):
    """
    Removes the given container using docker API.
    """
    with ContextDockerClient() as docker_client:
        for container in containers:
            try:
                docker_client.containers.get(container).remove(force=True)
            except Exception as e:
                logging.error(e)



from typing import Optional

from exasol_integration_test_docker_environment.lib.data.database_info import DatabaseInfo
from exasol_integration_test_docker_environment.lib.data.docker_network_info import DockerNetworkInfo
from exasol_integration_test_docker_environment.lib.data.docker_volume_info import DockerVolumeInfo
from exasol_integration_test_docker_environment.lib.data.environment_type import EnvironmentType
from exasol_integration_test_docker_environment.lib.test_environment.abstract_spawn_test_environment import \
    AbstractSpawnTestEnvironment
from exasol_integration_test_docker_environment.lib.test_environment.database_waiters.wait_for_external_database import \
    WaitForTestExternalDatabase
from exasol_integration_test_docker_environment.lib.test_environment.parameter.external_test_environment_parameter import \
    ExternalDatabaseXMLRPCParameter, ExternalDatabaseHostParameter
from exasol_integration_test_docker_environment.lib.test_environment.prepare_network_for_test_environment import \
    PrepareDockerNetworkForTestEnvironment
from exasol_integration_test_docker_environment.lib.test_environment.setup_external_database_host import \
    SetupExternalDatabaseHost


class SpawnTestEnvironmentWithExternalDB(AbstractSpawnTestEnvironment,
                                         ExternalDatabaseXMLRPCParameter,
                                         ExternalDatabaseHostParameter):

    def get_environment_type(self):
        return EnvironmentType.external_db

    def create_network_task(self, attempt: int):
        return \
            self.create_child_task_with_common_params(
                PrepareDockerNetworkForTestEnvironment,
                reuse=self.reuse_test_container,
                attempt=attempt,
                test_container_name=self.test_container_name,
                network_name=self.network_name,
                no_cleanup_after_success=self.reuse_test_container,
                no_cleanup_after_failure=self.reuse_test_container
            )

    def create_spawn_database_task(self, network_info: DockerNetworkInfo,
                                   certificate_volume_info: Optional[DockerVolumeInfo], attempt: int):
        if certificate_volume_info is not None:
            raise ValueError("Certificate volume must be None when using external database")

        return \
            self.create_child_task_with_common_params(
                SetupExternalDatabaseHost,
                network_info=network_info,
                attempt=attempt
            )

    def create_wait_for_database_task(self,
                                      attempt: int,
                                      database_info: DatabaseInfo):
        return \
            self.create_child_task_with_common_params(
                WaitForTestExternalDatabase,
                database_info=database_info,
                attempt=attempt)



from typing import List

import luigi
import netaddr
import pkg_resources
from docker.models.containers import Container
from docker.transport import unixconn

from exasol_integration_test_docker_environment.lib import PACKAGE_NAME
from exasol_integration_test_docker_environment.lib.base.docker_base_task import DockerBaseTask
from exasol_integration_test_docker_environment.lib.base.json_pickle_parameter import JsonPickleParameter
from exasol_integration_test_docker_environment.lib.data.container_info import ContainerInfo
from exasol_integration_test_docker_environment.lib.data.docker_network_info import DockerNetworkInfo
from exasol_integration_test_docker_environment.lib.docker.images.image_info import ImageState, ImageInfo
from exasol_integration_test_docker_environment.lib.test_environment.analyze_test_container import \
    DockerTestContainerBuild
from exasol_integration_test_docker_environment.lib.test_environment.docker_container_copy import \
    copy_script_to_container
from exasol_integration_test_docker_environment.lib.test_environment.parameter.test_container_parameter import \
    TestContainerParameter


class SpawnTestContainer(DockerBaseTask, TestContainerParameter):
    environment_name = luigi.Parameter()
    test_container_name = luigi.Parameter()
    network_info = JsonPickleParameter(
        DockerNetworkInfo, significant=False)  # type: DockerNetworkInfo
    ip_address_index_in_subnet = luigi.IntParameter(significant=False)
    attempt = luigi.IntParameter(1)
    reuse_test_container = luigi.BoolParameter(False, significant=False)
    no_test_container_cleanup_after_success = luigi.BoolParameter(False, significant=False)
    no_test_container_cleanup_after_failure = luigi.BoolParameter(False, significant=False)
    docker_runtime = luigi.OptionalParameter(None, significant=False)
    certificate_volume_name = luigi.OptionalParameter(None, significant=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.ip_address_index_in_subnet < 0:
            raise Exception(
                "ip_address_index_in_subnet needs to be greater than 0 got %s"
                % self.ip_address_index_in_subnet)

    def register_required(self):
        self.test_container_image_future = \
            self.register_dependency(self.create_child_task(task_class=DockerTestContainerBuild,
                                                            test_container_content=self.test_container_content))

    def is_reuse_possible(self) -> bool:
        test_container_image_info = \
            self.get_values_from_futures(self.test_container_image_future)["test-container"]  # type: ImageInfo
        test_container = None
        with self._get_docker_client() as docker_client:
            try:
                test_container = docker_client.containers.get(self.test_container_name)
            except Exception as e:
                pass
            ret_val = self.network_info.reused and self.reuse_test_container and \
                      test_container is not None and \
                      test_container_image_info.get_target_complete_name() in test_container.image.tags and \
                      test_container_image_info.image_state == ImageState.USED_LOCAL.name

        return ret_val

    def run_task(self):
        subnet = netaddr.IPNetwork(self.network_info.subnet)
        ip_address = str(subnet[2 + self.ip_address_index_in_subnet])
        container_info = None

        if self.is_reuse_possible():
            container_info = self._try_to_reuse_test_container(ip_address, self.network_info)
        if container_info is None:
            container_info = self._create_test_container(ip_address, self.network_info)
        with self._get_docker_client() as docker_client:
            docker_client.containers.get(self.test_container_name)
        self._copy_runtime_targets()
        self.return_object(container_info)

    def _copy_runtime_targets(self):
        self.logger.info("Copy runtime targets in test container %s.", self.test_container_name)
        with self._get_docker_client() as docker_client:
            test_container = docker_client.containers.get(self.test_container_name)
            for runtime_mapping in self.test_container_content.runtime_mappings:
                if runtime_mapping.deployment_target is not None:
                    self.logger.warning(f"Copy runtime target {runtime_mapping.target} "
                                        f"in test container {self.test_container_name}.")
                    try:
                        test_container.exec_run(cmd=f"rm -r {runtime_mapping.deployment_target}")
                    except:
                        pass
                    test_container.exec_run(cmd=f"cp -r {runtime_mapping.target} {runtime_mapping.deployment_target}")

    def _try_to_reuse_test_container(self, ip_address: str,
                                     network_info: DockerNetworkInfo) -> ContainerInfo:
        self.logger.info("Try to reuse test container %s",
                         self.test_container_name)
        container_info = None
        try:
            network_aliases = self._get_network_aliases()
            container_info = self.create_container_info(ip_address, network_aliases, network_info)
        except Exception as e:
            self.logger.warning("Tried to reuse test container %s, but got Exeception %s. "
                                "Fallback to create new database.", self.test_container_name, e)
        return container_info

    def _create_test_container(self, ip_address,
                               network_info: DockerNetworkInfo) -> ContainerInfo:
        self._remove_container(self.test_container_name)
        self.logger.info(f"Creating new test container {self.test_container_name}")
        test_container_image_info = \
            self.get_values_from_futures(self.test_container_image_future)["test-container"]

        volumes = dict()
        for runtime_mapping in self.test_container_content.runtime_mappings:
            volumes[runtime_mapping.source.absolute()] = {
                "bind": runtime_mapping.target,
                "mode": "rw"
            }
        if self.certificate_volume_name is not None:
            volumes[self.certificate_volume_name] = {
                "bind": "/certificates",
                "mode": "ro"
            }

        with self._get_docker_client() as docker_client:
            docker_unix_sockets = [i for i in docker_client.api.adapters.values()
                                   if isinstance(i, unixconn.UnixHTTPAdapter)]
            if len(docker_unix_sockets) > 0:
                host_docker_socker_path = docker_unix_sockets[0].socket_path
                volumes[host_docker_socker_path] = {
                    "bind": "/var/run/docker.sock",
                    "mode": "rw"
                }
            test_container = \
                docker_client.containers.create(
                    image=test_container_image_info.get_target_complete_name(),
                    name=self.test_container_name,
                    network_mode=None,
                    command="sleep infinity",
                    detach=True,
                    volumes=volumes,
                    labels={"test_environment_name": self.environment_name, "container_type": "test_container"},
                    runtime=self.docker_runtime
                )
            docker_network = docker_client.networks.get(network_info.network_name)
            network_aliases = self._get_network_aliases()
            docker_network.connect(test_container, ipv4_address=ip_address, aliases=network_aliases)
            test_container.start()
            self.register_certificates(test_container)
            container_info = self.create_container_info(ip_address, network_aliases, network_info)
            return container_info

    def _get_network_aliases(self):
        network_aliases = ["test_container", self.test_container_name]
        return network_aliases

    def create_container_info(self, ip_address: str, network_aliases: List[str],
                              network_info: DockerNetworkInfo) -> ContainerInfo:
        with self._get_docker_client() as docker_client:
            test_container = docker_client.containers.get(self.test_container_name)
            if test_container.status != "running":
                raise Exception(f"Container {self.test_container_name} not running")
            container_info = ContainerInfo(container_name=self.test_container_name,
                                           ip_address=ip_address,
                                           network_aliases=network_aliases,
                                           network_info=network_info)
        return container_info

    def _get_export_directory(self):
        return self.get_values_from_future(self.export_directory_future)

    def _remove_container(self, container_name: str):
        try:
            with self._get_docker_client() as docker_client:
                container = docker_client.containers.get(container_name)
                container.remove(force=True)
                self.logger.info(f"Removed container: name: '{container_name}', id: '{container.short_id}'")
        except Exception as e:
            pass

    def register_certificates(self, test_container: Container):
        if self.certificate_volume_name is not None:
            script_name = "install_root_certificate.sh"
            script_str = pkg_resources.resource_string(
                PACKAGE_NAME,
                f"certificate_resources/{script_name}")  # type: bytes

            script_location_in_container = f"scripts/{script_name}"
            copy_script_to_container(script_str.decode("UTF-8"), script_location_in_container, test_container)

            exit_code, output = test_container.exec_run(f"bash {script_location_in_container}")
            if exit_code != 0:
                raise RuntimeError(f"Error installing certificates:{output.decode('utf-8')}")

    def cleanup_task(self, success: bool):
        if (success and not self.no_test_container_cleanup_after_success) or \
                (not success and not self.no_test_container_cleanup_after_failure):
            try:
                self.logger.info(f"Cleaning up container %s", self.test_container_name)
                self._remove_container(self.test_container_name)
            except Exception as e:
                self.logger.error(f"Error during removing container %s: %s", self.test_container_name, e)



from typing import Set, Dict

from exasol_integration_test_docker_environment.lib.config.build_config import build_config
from exasol_integration_test_docker_environment.lib.config.docker_config import target_docker_repository_config, \
    source_docker_repository_config
from exasol_integration_test_docker_environment.lib.docker.images.create.docker_build_base import DockerBuildBase
from exasol_integration_test_docker_environment.lib.docker.images.create.docker_image_analyze_task import \
    DockerAnalyzeImageTask
from exasol_integration_test_docker_environment.lib.docker.images.push.docker_push_parameter import DockerPushParameter
from exasol_integration_test_docker_environment.lib.docker.images.push.push_task_creator_for_build_tasks import \
    PushTaskCreatorFromBuildTasks
from exasol_integration_test_docker_environment.lib.test_environment.parameter.test_container_parameter import \
    TestContainerParameter


class AnalyzeTestContainer(DockerAnalyzeImageTask, TestContainerParameter):

    def get_target_repository_name(self) -> str:
        return f"""{target_docker_repository_config().repository_name}"""

    def get_source_repository_name(self) -> str:
        return f"""{source_docker_repository_config().repository_name}"""

    def get_source_image_tag(self):
        if source_docker_repository_config().tag_prefix != "":
            return f"{source_docker_repository_config().tag_prefix}_db-test-container"
        else:
            return f"db-test-container"

    def get_target_image_tag(self):
        if target_docker_repository_config().tag_prefix != "":
            return f"{target_docker_repository_config().tag_prefix}_db-test-container"
        else:
            return f"db-test-container"

    def get_mapping_of_build_files_and_directories(self):
        return {mapping.target: str(mapping.source) for mapping
                in self.test_container_content.build_files_and_directories}

    def get_dockerfile(self):
        return str(self.test_container_content.docker_file)

    def is_rebuild_requested(self) -> bool:
        config = build_config()
        return config.force_rebuild


class DockerTestContainerBuildBase(DockerBuildBase, TestContainerParameter):

    def get_goal_class_map(self) -> Dict[str, DockerAnalyzeImageTask]:
        goal_class_map = {"test-container": self.create_child_task(task_class=AnalyzeTestContainer,
                                                                   test_container_content=self.test_container_content)}
        return goal_class_map

    def get_default_goals(self) -> Set[str]:
        goals = {"test-container"}
        return goals

    def get_goals(self):
        goals = {"test-container"}
        return goals


class DockerTestContainerBuild(DockerTestContainerBuildBase):

    def run_task(self):
        build_tasks = self.create_build_tasks(False)
        image_infos_futures = yield from self.run_dependencies(build_tasks)
        image_infos = self.get_values_from_futures(image_infos_futures)
        self.return_object(image_infos)


class DockerTestContainerPush(DockerTestContainerBuildBase, DockerPushParameter):

    def run_task(self):
        build_tasks = self.create_build_tasks(shortcut_build=not self.push_all)
        push_task_creator = PushTaskCreatorFromBuildTasks(self)
        push_tasks = push_task_creator.create_tasks_for_build_tasks(build_tasks)
        image_info_future = yield from self.run_dependencies(push_tasks)
        image_info = self.get_values_from_futures(image_info_future)
        self.return_object(image_info)



from docker.models.containers import Container
import io
import tarfile
import time

class DockerContainerCopy:

    def __init__(self, container:Container):
        super().__init__()
        self.open = True
        self.file_like_object = io.BytesIO()
        self.tar = tarfile.open(fileobj=self.file_like_object, mode="x")
        self.container = container

    def __del__(self):
        if self.open:
            self.tar.close()
            self.open = False
            self.tar = None
            self.file_like_object = None


    def is_open_or_raise(self):
        if not self.open:
            raise Exception("DockerContainerCopy not open")

    def add_string_to_file(self, name: str, string: str):
        self.is_open_or_raise()
        encoded = string.encode('utf-8')
        bytes_io = io.BytesIO(encoded)
        tar_info = tarfile.TarInfo(name=name)
        tar_info.mtime = time.time()
        tar_info.size = len(encoded)
        self.tar.addfile(tarinfo=tar_info, fileobj=bytes_io)

    def add_file(self, host_path:str, path_in_tar:str):
        self.is_open_or_raise()
        self.tar.add(host_path, path_in_tar)

    def copy(self, path_in_container:str):
        self.is_open_or_raise()
        self.tar.close()
        self.open = False
        self.tar = None
        self.container.put_archive(path_in_container, self.file_like_object.getbuffer().tobytes())
        self.file_like_object = None


def copy_script_to_container(script: str, script_path: str, container: Container):
    """
    Copy a script, given as string, to the container under specified location, relative to "/"
    """
    docker_container_copy = DockerContainerCopy(container)
    docker_container_copy.add_string_to_file(script_path, script)
    docker_container_copy.copy("/")



import docker
import luigi

from exasol_integration_test_docker_environment.lib.base.docker_base_task import DockerBaseTask
from exasol_integration_test_docker_environment.lib.data.docker_network_info import DockerNetworkInfo


class PrepareDockerNetworkForTestEnvironment(DockerBaseTask):
    environment_name = luigi.Parameter()
    network_name = luigi.Parameter()
    test_container_name = luigi.Parameter(significant=False)
    db_container_name = luigi.OptionalParameter(None, significant=False)
    reuse = luigi.BoolParameter(False, significant=False)
    no_cleanup_after_success = luigi.BoolParameter(False, significant=False)
    no_cleanup_after_failure = luigi.BoolParameter(False, significant=False)
    attempt = luigi.IntParameter(-1)

    def run_task(self):
        self.network_info = None
        if self.reuse:
            self.logger.info("Try to reuse network %s", self.network_name)
            try:
                self.network_info = self.get_network_info(reused=True)
            except Exception as e:
                self.logger.warning("Tried to reuse network %s, but got Exeception %s. "
                                    "Fallback to create new network.", self.network_name, e)
        if self.network_info is None:
            self.network_info = self.create_docker_network()
        self.return_object(self.network_info)

    def get_network_info(self, reused: bool):
        with self._get_docker_client() as docker_client:
            network_properties = docker_client.api.inspect_network(self.network_name)
            network_config = network_properties["IPAM"]["Config"][0]
            return DockerNetworkInfo(network_name=self.network_name, subnet=network_config["Subnet"],
                                     gateway=network_config["Gateway"], reused=reused)

    def create_docker_network(self) -> DockerNetworkInfo:
        self.remove_container(self.test_container_name)
        if self.db_container_name is not None:
            self.remove_container(self.db_container_name)
        self.remove_network(self.network_name)
        with self._get_docker_client() as docker_client:
            network = docker_client.networks.create(
                name=self.network_name,
                driver="bridge",
            )
            network_info = self.get_network_info(reused=False)
            subnet = network_info.subnet
            gateway = network_info.gateway
            ipam_pool = docker.types.IPAMPool(
                subnet=subnet,
                gateway=gateway
            )
            ipam_config = docker.types.IPAMConfig(
                pool_configs=[ipam_pool]
            )
            self.remove_network(self.network_name)  # TODO race condition possible, add retry
            network = docker_client.networks.create(
                name=self.network_name,
                driver="bridge",
                ipam=ipam_config
            )
        return network_info

    def remove_network(self, network_name):
        try:
            with self._get_docker_client() as docker_client:
                docker_client.networks.get(network_name).remove()
                self.logger.info("Removed network %s", network_name)
        except docker.errors.NotFound:
            pass

    def remove_container(self, container_name: str):
        try:
            with self._get_docker_client() as docker_client:
                container = docker_client.containers.get(container_name)
                container.remove(force=True)
                self.logger.info("Removed container %s", container_name)
        except docker.errors.NotFound:
            pass

    def cleanup_task(self, success):
        if (success and not self.no_cleanup_after_success) or \
            (not success and not self.no_cleanup_after_failure):
            try:
                self.logger.info(f"Cleaning up container %s:",self.test_container_name)
                self.remove_container(self.test_container_name)
            except Exception as e:
                self.logger.error(f"Error during removing container %s: %s:", self.test_container_name,e)

            if self.db_container_name is not None:
                try:
                    self.logger.info(f"Cleaning up container %s",self.db_container_name)
                    self.remove_container(self.db_container_name)
                except Exception as e:
                    self.logger.error(f"Error during removing container %s: %s: ",self.db_container_name,e)

            try:
                self.logger.info(f"Cleaning up dpcker network %s", self.network_name)
                self.remove_network(self.network_name)
            except Exception as e:
                self.logger.error(f"Error during removing container %s: %s",self.network_name,e)



import importlib

import luigi

from exasol_integration_test_docker_environment.lib.base.json_pickle_parameter import JsonPickleParameter
from exasol_integration_test_docker_environment.lib.docker.images.required_task_info import RequiredTaskInfo
from exasol_integration_test_docker_environment.lib.docker.images.save.docker_image_save_base_task import \
    DockerSaveImageBaseTask


class DockerSaveImageTask(DockerSaveImageBaseTask):
    # We need to create the DockerCreateImageTask for DockerSaveImageTask dynamically,
    # because we want to save as soon as possible after an image was build and
    # don't want to wait for the save finishing before starting to build depended images,
    # but we also need to create a DockerSaveImageTask for each DockerCreateImageTask of a goal

    required_task_info = JsonPickleParameter(RequiredTaskInfo,
                                             visibility=luigi.parameter.ParameterVisibility.HIDDEN,
                                             significant=True)  # type: RequiredTaskInfo

    def get_docker_image_task(self):
        module = importlib.import_module(self.required_task_info.module_name)
        class_ = getattr(module, self.required_task_info.class_name)
        instance = self.create_child_task(class_, **self.required_task_info.params)
        return instance



import traceback
from pathlib import Path
from typing import List, Dict
from collections import OrderedDict
from exasol_integration_test_docker_environment.lib.base.timeable_base_task import TimeableBaseTask


class StoppingFurtherExecution(Exception):
    pass


class StoppableBaseTask(TimeableBaseTask):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.failed_target = Path(super()._get_output_path_for_job(), "TASK_FAILED")

    def get_failure_log_path(self) -> Path:
        path = Path(self.get_output_path(), "failure")
        return path

    def run(self):
        try:
            self.fail_if_any_task_failed()
            task_generator = super().run()
            if task_generator is not None:
                yield from task_generator
        except Exception as exception:
            exception_tb = traceback.format_exc()
            self.handle_failure(exception, exception_tb)
            raise exception

    def fail_if_any_task_failed(self):
        if self.failed_target.exists():
            with self.failed_target.open("r") as f:
                failed_task = f.read()
            self.logger.error("Task %s failed. Stopping further execution." % failed_task)
            raise StoppingFurtherExecution("Task %s failed. Stopping further execution." % failed_task)

    def handle_failure(self, exception, exception_tb):
        if not isinstance(exception, StoppingFurtherExecution):
            with self.get_failure_log_path().open("w") as f:
                f.write("%s" % exception_tb)
            if not self.failed_target.exists():
                with self.failed_target.open("w") as f:
                    f.write("%s" % self.task_id)

    def collect_failures(self) -> Dict[str,None]:
        failures = OrderedDict()
        failures.update(self.collect_failures_of_child_tasks())
        if self.get_failure_log_path().exists():
            with self.get_failure_log_path().open("r") as f:
                exception = f.read().strip()
                prefix = '    '
                formatted_exception = prefix + prefix.join(exception.splitlines(True))
                failure_message = "- %s:\n%s" % (self.task_id, formatted_exception)
                failures[failure_message] = None

        return failures

    def collect_failures_of_child_tasks(self) -> Dict[str,None]:
        failures_of_child_tasks = OrderedDict()
        if self._run_dependencies_target.exists():
            _run_dependencies_tasks_from_target = self._run_dependencies_target.read()
        else:
            _run_dependencies_tasks_from_target = []
        _run_dependencies_tasks = self._run_dependencies_tasks + _run_dependencies_tasks_from_target
        reversed_run_dependencies_task_list = list(_run_dependencies_tasks)
        reversed_run_dependencies_task_list.reverse()
        for task in reversed_run_dependencies_task_list:
            if isinstance(task, StoppableBaseTask):
                failures_of_child_tasks.update(task.collect_failures())

        reversed_registered_task_list = list(self._registered_tasks)
        reversed_registered_task_list.reverse()
        for task in reversed_registered_task_list:
            if isinstance(task, StoppableBaseTask):
                failures_of_child_tasks.update(task.collect_failures())
        return failures_of_child_tasks



"""
The doctor module provides functionality to check the health of the `exasol_integration_test_docker_environment`
package and also provide help to find potential fixes.
"""
import sys
from enum import Enum
from typing import Iterator

import docker
from docker.errors import DockerException

SUPPORTED_PLATFORMS = ["linux", "darwin"]


class ErrorCodes(Enum):
    """The equivalent of ICD-10 codes this doctor is using"""

    Unknown = "Unknown issue"
    UnixSocketNotAvailable = "Could not find unix socket to connect to"
    TargetPlatformNotSupported = "The platform you are running on is not supported."


def recommend_mitigation(error_code) -> str:
    """Get treatment advice based on the error_code"""
    return {
        ErrorCodes.Unknown: "You are sick but this symptoms are unknown, please contact the maintainer.",
        ErrorCodes.UnixSocketNotAvailable: "Make sure your DOCKER_HOST environment variable is configured correctly.",
        ErrorCodes.TargetPlatformNotSupported: f"Make sure you are using one of the following platforms: {SUPPORTED_PLATFORMS}.",
    }[error_code]


def diagnose_docker_daemon_not_available() -> Iterator[ErrorCodes]:
    """Diagnose reasons why docker deamon is not available"""

    def _is_unix_socket_issue(message: str) -> bool:
        return "FileNotFoundError(2, 'No such file or directory')" in message

    errors = set()
    try:
        _docker = docker.from_env()
    except DockerException as ex:
        msg = f"{ex}"
        if _is_unix_socket_issue(msg):
            errors.add(ErrorCodes.UnixSocketNotAvailable)
        if len(errors) == 0:
            errors.add(ErrorCodes.Unknown)
    return errors


def is_docker_daemon_available() -> bool:
    """
    Checks weather or not the docker daemon is available.
    """
    try:
        _docker = docker.from_env()
    except DockerException:
        return False
    return True


def is_supported_platform() -> bool:
    """
    Checks weather or not the current platform is supported.
    """
    return sys.platform in SUPPORTED_PLATFORMS


def health_checkup() -> Iterator[ErrorCodes]:
    """
    Runs all known examinations

    return an iterator of error codes specifying which problems have been identified.
    """
    examinations = [
        (is_docker_daemon_available, diagnose_docker_daemon_not_available),
        (is_supported_platform, lambda: ErrorCodes.TargetPlatformNotSupported),
    ]
    for is_fine, diagnosis in examinations:
        if not is_fine():
            for error_code in diagnosis():
                yield error_code



import time
from datetime import datetime

from docker.models.containers import Container

from exasol_integration_test_docker_environment.lib.test_environment.database_setup.bucketfs_sync_checker import \
    BucketFSSyncChecker


class DockerDBLogBasedBucketFSSyncChecker(BucketFSSyncChecker):

    def __init__(self, logger,
                 database_container: Container,
                 pattern_to_wait_for: str,
                 log_file_to_check: str,
                 bucketfs_write_password: str):
        self.logger = logger
        self.pattern_to_wait_for = pattern_to_wait_for
        self.log_file_to_check = log_file_to_check
        self.database_container = database_container
        self.bucketfs_write_password = bucketfs_write_password

    def prepare_upload(self):
        self.start_exit_code, self.start_output = self.find_pattern_in_logfile()

    def wait_for_bucketfs_sync(self):
        self.logger.info("wait for upload of file")

        ready = False
        while not ready:
            exit_code, output = self.find_pattern_in_logfile()
            if self.exit_code_changed(exit_code, self.start_exit_code) or \
                    self.found_new_log_line(exit_code, self.start_exit_code,
                                            self.start_output, output):
                ready = True
            time.sleep(1)

    def exit_code_changed(self, exit_code, start_exit_code):
        return exit_code == 0 and start_exit_code != 0

    def found_new_log_line(self, exit_code, start_exit_code,
                           start_output, output):
        return exit_code == 0 and start_exit_code == 0 and len(start_output) < len(output)

    def find_pattern_in_logfile(self):
        cmd = f"""grep '{self.pattern_to_wait_for}' {self.log_file_to_check}"""
        bash_cmd = f"""bash -c "{cmd}" """
        exit_code, output = \
            self.database_container.exec_run(bash_cmd)
        return exit_code, output

    def output_happened_after_start_time(self, output, start_time):
        time_str_from_output = " ".join(output.decode("utf-8").split(" ")[1:3])
        time_from_output = datetime.strptime(time_str_from_output, "%y%m%d %H:%M:%S")
        happened_after = time_from_output > start_time
        return happened_after



from typing import Callable, List


def find_images_by_tag(docker_client, condition: Callable[[str], bool]) -> List:
    images = docker_client.images.list()
    filter_images = [image for image in images
                     if image.tags is not None and len(image.tags) > 0 and
                     any([condition(tag) for tag in image.tags])]
    return filter_images



from typing import Tuple, Optional

from exasol_integration_test_docker_environment.lib.api.common import set_build_config, \
    set_docker_repository_config, run_task, generate_root_task, no_cli_function
from exasol_integration_test_docker_environment.cli.options.docker_repository_options import \
    DEFAULT_DOCKER_REPOSITORY_NAME
from exasol_integration_test_docker_environment.cli.options.system_options import DEFAULT_OUTPUT_DIRECTORY
from exasol_integration_test_docker_environment.lib.data.test_container_content_description import \
    TestContainerContentDescription
from exasol_integration_test_docker_environment.lib.docker.images.image_info import ImageInfo
from exasol_integration_test_docker_environment.lib.test_environment.analyze_test_container import AnalyzeTestContainer, \
    DockerTestContainerBuild


@no_cli_function
def build_test_container(
        test_container_content: TestContainerContentDescription,
        force_rebuild: bool = False,
        force_rebuild_from: Tuple[str, ...] = tuple(),
        force_pull: bool = False,
        output_directory: str = DEFAULT_OUTPUT_DIRECTORY,
        temporary_base_directory: str = "/tmp",
        log_build_context_content: bool = False,
        cache_directory: Optional[str] = None,
        build_name: Optional[str] = None,
        source_docker_repository_name: str = DEFAULT_DOCKER_REPOSITORY_NAME,
        source_docker_tag_prefix: str = '',
        source_docker_username: Optional[str] = None,
        source_docker_password: Optional[str] = None,
        target_docker_repository_name: str = DEFAULT_DOCKER_REPOSITORY_NAME,
        target_docker_tag_prefix: str = '',
        target_docker_username: Optional[str] = None,
        target_docker_password: Optional[str] = None,
        workers: int = 5,
        task_dependencies_dot_file: Optional[str] = None,
        log_level: Optional[str] = None,
        use_job_specific_log_file: bool = False
) -> ImageInfo:
    """
    This function builds all stages of the test container for the test environment.
    If stages are cached in a docker registry, the function is going to pull them,
    instead of building them.
    It returns the image info of the test-container.
    """
    set_build_config(force_rebuild,
                     force_rebuild_from,
                     force_pull,
                     log_build_context_content,
                     output_directory,
                     temporary_base_directory,
                     cache_directory,
                     build_name)
    # Use AnalyzeTestContainer to ensure that all luigi processes got it loaded
    analyze_task = AnalyzeTestContainer.__class__.__name__

    set_docker_repository_config(source_docker_password, source_docker_repository_name, source_docker_username,
                                 source_docker_tag_prefix, "source")
    set_docker_repository_config(target_docker_password, target_docker_repository_name, target_docker_username,
                                 target_docker_tag_prefix, "target")
    task_creator = lambda: generate_root_task(task_class=DockerTestContainerBuild,
                                              test_container_content=test_container_content)
    image_infos = run_task(task_creator, workers, task_dependencies_dot_file,
                           log_level=log_level,
                           use_job_specific_log_file=use_job_specific_log_file)
    return image_infos["test-container"]



from datetime import datetime
from pathlib import Path

from exasol_integration_test_docker_environment.lib.base.base_task import BaseTask


class TimeableBaseTask(BaseTask):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.write_creation_time()

    def _get_timers_path_for_job(self):
        return Path(super()._get_output_path_for_job(), "timers")

    def _get_timers_path(self):
        return Path(self._get_timers_path_for_job(), self.task_id)

    def _get_timers_state_path(self):
        path = Path(self._get_timers_path(), "state")
        path.mkdir(parents=True, exist_ok=True)
        return path

    def _get_timers_result_path(self):
        path = Path(self._get_timers_path(), "result")
        path.mkdir(parents=True, exist_ok=True)
        return path

    def _get_timers_creation_state_path(self):
        return Path(self._get_timers_state_path(), "creation")

    def _get_timers_first_run_state_path(self):
        return Path(self._get_timers_state_path(), "first_run")

    def _get_timers_runs_state_path(self):
        return Path(self._get_timers_state_path(), "runs")

    def _get_timers_creation_result_path(self):
        return Path(self._get_timers_result_path(), "time_since_creation")

    def _get_timers_first_run_result_path(self):
        return Path(self._get_timers_result_path(), "time_since_first_run")

    def _get_timers_runs_result_path(self):
        return Path(self._get_timers_result_path(), "total_time_of_runs")

    def write_creation_time(self):
        state_path = self._get_timers_creation_state_path()
        if not state_path.exists():
            with state_path.open("w") as f:
                f.write(str(datetime.now().timestamp()))

    def write_first_run_start_time(self, start_time):
        state_path = self._get_timers_first_run_state_path()
        if not state_path.exists():
            with state_path.open("w") as f:
                f.write(str(start_time.timestamp()))

    def write_runs_time(self, start_time):
        state_path = self._get_timers_runs_state_path()
        timedelta = datetime.now() - start_time
        with state_path.open("a") as f:
            f.write(str(timedelta.total_seconds()))
            f.write("\n")

    def on_success(self):
        super().on_success()
        now = datetime.now()
        self.log_time_since_first_run(now)
        self.log_time_since_creation(now)
        self.log_time_of_runs()

    def log_time_since_creation(self, now):
        state_path = self._get_timers_creation_state_path()
        if state_path.exists():
            with state_path.open("r") as f:
                start_time_str = f.read()
            start_time = datetime.fromtimestamp(float(start_time_str))
            timedelta = now - start_time
            self.logger.info(f"Time since creation {timedelta.total_seconds()} s")
            with self._get_timers_creation_result_path().open("w") as f:
                f.write(str(timedelta.total_seconds()))
                f.write("\n")

    def log_time_since_first_run(self, now):
        state_path = self._get_timers_first_run_state_path()
        if state_path.exists():
            with state_path.open("r") as f:
                start_time_str = f.read()
            start_time = datetime.fromtimestamp(float(start_time_str))
            timedelta = now - start_time
            self.logger.info("Time since first_run %s s", timedelta.total_seconds())
            with self._get_timers_first_run_result_path().open("w") as f:
                f.write(str(timedelta.total_seconds()))
                f.write("\n")

    def log_time_of_runs(self):
        state_path = self._get_timers_runs_state_path()
        if state_path.exists():
            with state_path.open("r") as f:
                total_runtime = self.calculate_total_runtime(f.readlines())
            self.logger.info("Total runtime of run method %s s", total_runtime)
            with self._get_timers_runs_result_path().open("w") as f:
                f.write(str(total_runtime))
                f.write("\n")

    def calculate_total_runtime(self, lines):
        total_runtime = 0
        for line in lines:
            seconds_of_run = float(line)
            total_runtime += seconds_of_run
        return total_runtime

    def run(self):
        start_time = datetime.now()
        self.write_first_run_start_time(start_time)
        try:
            task_generator = super().run()
            if task_generator is not None:
                yield from task_generator
        finally:
            self.write_runs_time(start_time)



from pathlib import Path

import luigi

from exasol_integration_test_docker_environment.abstract_method_exception import AbstractMethodException
from exasol_integration_test_docker_environment.lib.base.docker_base_task import DockerBaseTask
from exasol_integration_test_docker_environment.lib.base.still_running_logger import StillRunningLogger
from exasol_integration_test_docker_environment.lib.config.docker_config import target_docker_repository_config
from exasol_integration_test_docker_environment.lib.docker.images.image_info import ImageState, ImageInfo
from exasol_integration_test_docker_environment.lib.docker.images.push.push_log_handler import PushLogHandler


class DockerPushImageBaseTask(DockerBaseTask):
    image_name = luigi.Parameter()
    force_push = luigi.BoolParameter(False, visibility=luigi.parameter.ParameterVisibility.HIDDEN)

    def register_required(self):
        task = self.get_docker_image_task()
        self._image_info_future = self.register_dependency(task)

    def get_docker_image_task(self):
        raise AbstractMethodException()

    def run_task(self):
        image_info = self.get_values_from_future(self._image_info_future)
        was_build = image_info.image_state == ImageState.WAS_BUILD.name
        if was_build or self.force_push:
            self.logger.info("Push images")
            auth_config = {
                "username": target_docker_repository_config().username,
                "password": target_docker_repository_config().password
            }
            with self._get_docker_client() as docker_client:
                self.logger.info(f"Push images to repo={image_info.get_target_complete_name()}, "
                                 f"tag={image_info.get_target_complete_tag()}")
                generator = docker_client.images.push(repository=image_info.get_target_complete_name(),
                                               tag=image_info.get_target_complete_tag(),
                                               auth_config=auth_config,
                                               stream=True)
                self._handle_output(generator, image_info)
        self.return_object(image_info)

    def _handle_output(self, output_generator, image_info: ImageInfo):
        log_file_path = Path(self.get_log_path(), "push.log")
        with PushLogHandler(log_file_path, self.logger, image_info) as log_hanlder:
            still_running_logger = StillRunningLogger(
                self.logger, "push image %s" % image_info.get_target_complete_name())
            for log_line in output_generator:
                still_running_logger.log()
                log_hanlder.handle_log_lines(log_line)



from exasol_integration_test_docker_environment.lib.base.info import Info


class DockerVolumeInfo(Info):

    def __str__(self):
        return f'DockerVolumeInfo: \n' \
               f'volume_name="{self.volume_name}"\n' \
               f'reused={self.reused}\n' \
               f'mount_point={self.mount_point}'

    def __init__(self, volume_name: str, mount_point: str, reused: bool = False):
        self.volume_name = volume_name
        self.reused = reused
        self.mount_point = mount_point



from typing import Optional, Tuple

import click

from exasol_integration_test_docker_environment.cli.cli import cli
from exasol_integration_test_docker_environment.cli.options.docker_repository_options import (
    docker_repository_options,
)
from exasol_integration_test_docker_environment.cli.options.system_options import (
    output_directory_option,
    system_options,
    tempory_base_directory_option, luigi_logging_options,
)
from exasol_integration_test_docker_environment.cli.options.test_environment_options import (
    docker_db_options,
)
from exasol_integration_test_docker_environment.cli.termination_handler import (
    TerminationHandler,
)
from exasol_integration_test_docker_environment.lib import api
from exasol_integration_test_docker_environment.lib.api.api_errors import (
    ArgumentConstraintError,
)
from exasol_integration_test_docker_environment.lib.api.common import add_options


@cli.command()
@click.option(
    "--environment-name",
    type=str,
    required=True,
    help="Name of the docker environment. This name gets used as suffix for the container db_container_<name> and test_container_<name>",
)
@click.option(
    "--database-port-forward",
    type=int,
    default=None,
    show_default=True,
    help="Host port to which the database port gets forwarded",
)
@click.option(
    "--bucketfs-port-forward",
    type=int,
    default=None,
    show_default=True,
    help="Host port to which the BucketFS port gets forwarded",
)
@click.option(
    "--db-mem-size",
    type=str,
    default="2 GiB",
    show_default=True,
    help="The main memory used by the database. Format <number> <unit>, e.g. 1 GiB. The minimum size is 1 GB, below that the database will not start.",
)
@click.option(
    "--db-disk-size",
    type=str,
    default="2 GiB",
    show_default=True,
    help="The disk size available for the database. Format <number> <unit>, e.g. 1 GiB. The minimum size is 100 MiB. However, the setup creates volume files with at least 2 GB larger size, because the database needs at least so much more disk.",
)
@click.option(
    "--nameserver",
    type=str,
    default=[],
    multiple=True,
    help="Add a nameserver to the list of DNS nameservers which the docker-db should use for resolving domain names. You can repeat this option to add further nameservers.",
)
@click.option(
    "--docker-runtime",
    type=str,
    default=None,
    show_default=True,
    help="The docker runtime used to start all containers",
)
@add_options(docker_db_options)
@add_options(docker_repository_options)
@add_options([output_directory_option])
@add_options([tempory_base_directory_option])
@add_options(system_options)
@add_options(luigi_logging_options)
def spawn_test_environment(
        environment_name: str,
        database_port_forward: Optional[int],
        bucketfs_port_forward: Optional[int],
        db_mem_size: str,
        db_disk_size: str,
        nameserver: Tuple[str, ...],
        docker_runtime: Optional[str],
        docker_db_image_version: str,
        docker_db_image_name: str,
        db_os_access: Optional[str],
        create_certificates: bool,
        additional_db_parameter: Tuple[str, ...],
        source_docker_repository_name: str,
        source_docker_tag_prefix: str,
        source_docker_username: Optional[str],
        source_docker_password: Optional[str],
        target_docker_repository_name: str,
        target_docker_tag_prefix: str,
        target_docker_username: Optional[str],
        target_docker_password: Optional[str],
        output_directory: str,
        temporary_base_directory: str,
        workers: int,
        task_dependencies_dot_file: Optional[str],
        log_level: Optional[str],
        use_job_specific_log_file: bool
):
    """
    This command spawns a test environment with a docker-db container and a connected test-container.
    The test-container is reachable by the database for output redirects of UDFs.
    """
    with TerminationHandler():
        try:
            api.spawn_test_environment(
                environment_name,
                database_port_forward,
                bucketfs_port_forward,
                db_mem_size,
                db_disk_size,
                nameserver,
                docker_runtime,
                docker_db_image_version,
                docker_db_image_name,
                db_os_access,
                create_certificates,
                additional_db_parameter,
                source_docker_repository_name,
                source_docker_tag_prefix,
                source_docker_username,
                source_docker_password,
                target_docker_repository_name,
                target_docker_tag_prefix,
                target_docker_username,
                target_docker_password,
                output_directory,
                temporary_base_directory,
                workers,
                task_dependencies_dot_file,
                log_level=log_level,
                use_job_specific_log_file=use_job_specific_log_file,
            )
        except ArgumentConstraintError as e:
            handle_wrong_argument_error(*e.args)


def handle_wrong_argument_error(argument_name, message):
    formatted = f"--{argument_name}".replace("_", "-")
    print(f"{formatted}: {message}")
    ctx = click.get_current_context()
    click.echo(ctx.get_help())
    exit(1)



from pathlib import Path
from typing import Dict, Any

import luigi

from exasol_integration_test_docker_environment.lib.base.dependency_logger_base_task import DependencyLoggerBaseTask
from exasol_integration_test_docker_environment.lib.base.docker_base_task import DockerBaseTask


class FlavorsBaseTask(DependencyLoggerBaseTask):
    flavor_paths = luigi.ListParameter()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for flavor_path in self.flavor_paths:
            if not Path(flavor_path).is_dir():
                raise OSError("Flavor path %s not a directory." % flavor_path)

    def create_tasks_for_flavors_with_common_params(self, cls, **kwargs) -> Dict[str, Any]:
        return {flavor_path: self._create_task_for_with_common_params(cls, flavor_path, kwargs)
                for flavor_path in self.flavor_paths}

    def _create_task_for_with_common_params(self, cls, flavor_path, kwargs):
        params = {**kwargs, "flavor_path": flavor_path}
        task = self.create_child_task_with_common_params(cls, **params)
        return task


class FlavorBaseTask(DockerBaseTask):
    flavor_path = luigi.Parameter()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not Path(self.flavor_path).is_dir():
            raise OSError("Flavor path %s not a directory." % self.flavor_path)

    def get_flavor_name(self):
        path = Path(self.flavor_path)
        flavor_name = path.name
        return flavor_name



from exasol_integration_test_docker_environment.lib.config.log_config import log_config


class AbstractLogHandler:

    def __init__(self, log_file_path, logger):
        self._log_file_path = log_file_path
        self._logger = logger
        self._complete_log = []
        self._error_message = None
        self._log_config = log_config()

    def __enter__(self):
        if self._log_file_path is not None:
            self._log_file = self._log_file_path.open("w")
        return self

    def __exit__(self, type, value, traceback):
        if self._log_file_path is not None:
            self._log_file.close()
        self.finish()

    def handle_log_lines(self, log_lines, error: bool = False):
        log_lines = log_lines.decode("utf-8")
        log_lines = log_lines.strip('\r\n')
        result = []
        for log_line in log_lines.split("\n"):
            log_line = log_line.strip('\r\n')
            result.append(self.handle_log_line(log_line, error))
        return result

    def handle_log_line(self, log_line, error: bool = False):
        pass

    def finish(self):
        pass



import jsonpickle
from luigi.parameter import Parameter, ParameterVisibility, _no_value


class JsonPickleParameter(Parameter):

    def __init__(self, cls, default=_no_value, is_global=False, significant=True, description=None,
                 config_path=None, positional=True, always_in_help=False, batch_method=None, is_optional=False,
                 visibility=ParameterVisibility.PUBLIC):
        super().__init__(default, is_global, significant, description,
                         config_path, positional, always_in_help,
                         batch_method, visibility)
        self.cls = cls
        self._is_optional = is_optional

    def parse(self, s):
        jsonpickle.set_preferred_backend('simplejson')
        loaded_object = jsonpickle.decode(s)
        if self._is_optional and loaded_object is None:
            return None
        elif not isinstance(loaded_object, self.cls):
            raise TypeError("Type %s of loaded object does not match %s" % (type(loaded_object), self.cls))
        return loaded_object

    def serialize(self, x):
        jsonpickle.set_preferred_backend('simplejson')
        return jsonpickle.encode(x)



import math
from typing import Tuple

import docker
import humanfriendly
import luigi
import netaddr
import pkg_resources
from docker.models.containers import Container
from docker.models.volumes import Volume
from jinja2 import Template

from exasol_integration_test_docker_environment.lib import PACKAGE_NAME
from exasol_integration_test_docker_environment.lib.base.docker_base_task import DockerBaseTask
from exasol_integration_test_docker_environment.lib.base.json_pickle_parameter import JsonPickleParameter
from exasol_integration_test_docker_environment.lib.base.still_running_logger import StillRunningLogger
from exasol_integration_test_docker_environment.lib.data.container_info import ContainerInfo
from exasol_integration_test_docker_environment.lib.data.database_info import DatabaseInfo
from exasol_integration_test_docker_environment.lib.data.docker_network_info import DockerNetworkInfo
from exasol_integration_test_docker_environment.lib.docker.images.create.utils.pull_log_handler import PullLogHandler
from exasol_integration_test_docker_environment.lib.docker.images.image_info import ImageInfo
from exasol_integration_test_docker_environment.lib.test_environment.db_version import DbVersion
from exasol_integration_test_docker_environment.lib.test_environment.docker_container_copy import DockerContainerCopy
from exasol_integration_test_docker_environment.lib \
    .test_environment.parameter \
    .docker_db_test_environment_parameter import (
        DbOsAccess,
        DockerDBTestEnvironmentParameter,
)
from exasol_integration_test_docker_environment.lib.base.ssh_access import (
    SshKeyCache,
    SshKey,
)


BUCKETFS_PORT = "6583"
DB_PORT = "8888"
CERTIFICATES_MOUNT_DIR = "/certificates"
CERTIFICATES_DEFAULT_DIR = "/exa/etc/ssl/"


class SpawnTestDockerDatabase(DockerBaseTask, DockerDBTestEnvironmentParameter):
    environment_name = luigi.Parameter()  # type: str
    db_container_name = luigi.Parameter()  # type: str
    attempt = luigi.IntParameter(1)  # type: int
    network_info = JsonPickleParameter(DockerNetworkInfo, significant=False)  # type: DockerNetworkInfo
    ip_address_index_in_subnet = luigi.IntParameter(significant=False)  # type: int
    docker_runtime = luigi.OptionalParameter(None, significant=False)  # type: str
    certificate_volume_name = luigi.OptionalParameter(None, significant=False)
    additional_db_parameter = luigi.ListParameter()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.ip_address_index_in_subnet < 0:
            raise Exception(
                "ip_address_index_in_subnet needs to be greater than 0 got %s"
                % self.ip_address_index_in_subnet)

        self.db_version = DbVersion.from_db_version_str(self.docker_db_image_version)
        self.docker_db_config_resource_name = f"docker_db_config/{self.db_version}"

    def run_task(self):
        subnet = netaddr.IPNetwork(self.network_info.subnet)
        db_ip_address = str(subnet[2 + self.ip_address_index_in_subnet])
        db_private_network = "{ip}/{prefix}".format(ip=db_ip_address, prefix=subnet.prefixlen)
        database_info = None
        if self.network_info.reused:
            database_info = self._try_to_reuse_database(db_ip_address)
        if database_info is None:
            database_info = self._create_database_container(db_ip_address, db_private_network)
        self.return_object(database_info)

    def _try_to_reuse_database(self, db_ip_address: str) -> DatabaseInfo:
        self.logger.info("Try to reuse database container %s",
                         self.db_container_name)
        database_info = None
        try:
            database_info = self._create_database_info(db_ip_address=db_ip_address, reused=True)
        except Exception as e:
            self.logger.warning("Tried to reuse database container %s, but got Exeception %s. "
                                "Fallback to create new database.", self.db_container_name, e)
        return database_info

    def _handle_output(self, output_generator, image_info: ImageInfo):
        log_file_path = self.get_log_path().joinpath("pull_docker_db_image.log")
        with PullLogHandler(log_file_path, self.logger, image_info) as log_handler:
            still_running_logger = StillRunningLogger(
                self.logger, "pull image %s" % image_info.get_source_complete_name())
            for log_line in output_generator:
                still_running_logger.log()
                log_handler.handle_log_lines(log_line)

    def _enable_ssh_access(self, container: Container):
        sshkey = SshKey.from_cache()
        copy = DockerContainerCopy(container)
        content = sshkey.public_key_as_string("itde-ssh-access")
        copy.add_string_to_file(".ssh/authorized_keys", content)
        copy.copy("/root/")

    def _create_database_container(self, db_ip_address: str, db_private_network: str):
        self.logger.info("Starting database container %s", self.db_container_name)
        with self._get_docker_client() as docker_client:
            try:
                docker_client.containers.get(self.db_container_name).remove(force=True, v=True)
            except:
                pass
            docker_db_image_info = self._pull_docker_db_images_if_necessary()
            db_volume = self._prepare_db_volume(docker_client, db_private_network, docker_db_image_info)
            ports = {}
            if self.database_port_forward is not None:
                ports[f"{DB_PORT}/tcp"] = ('0.0.0.0', int(self.database_port_forward))
            if self.bucketfs_port_forward is not None:
                ports[f"{BUCKETFS_PORT}/tcp"] = ('0.0.0.0', int(self.bucketfs_port_forward))
            volumes = {db_volume.name: {"bind": "/exa", "mode": "rw"}}
            if self.certificate_volume_name is not None:
                volumes[self.certificate_volume_name] = {"bind": CERTIFICATES_MOUNT_DIR, "mode": "ro"}
            db_container = \
                docker_client.containers.create(
                    image="%s" % (docker_db_image_info.get_source_complete_name()),
                    name=self.db_container_name,
                    detach=True,
                    privileged=True,
                    volumes=volumes,
                    network_mode=None,
                    ports=ports,
                    runtime=self.docker_runtime
                )
            if self.db_os_access == DbOsAccess.SSH:
                self._enable_ssh_access(db_container)
            docker_network = docker_client.networks.get(self.network_info.network_name)
            network_aliases = self._get_network_aliases()
            docker_network.connect(db_container, ipv4_address=db_ip_address, aliases=network_aliases)
            db_container.start()
            database_info = self._create_database_info(db_ip_address=db_ip_address, reused=False)
            return database_info

    def _get_network_aliases(self):
        network_aliases = ["exasol_test_database", "exasol-test-database", self.db_container_name]
        return network_aliases

    def _create_database_info(self, db_ip_address: str, reused: bool) -> DatabaseInfo:
        with self._get_docker_client() as docker_client:
            db_container = docker_client.containers.get(self.db_container_name)
            if db_container.status != "running":
                raise Exception(f"Container {self.db_container_name} not running")
            network_aliases = self._get_network_aliases()
            container_info = \
                ContainerInfo(container_name=self.db_container_name,
                              ip_address=db_ip_address,
                              network_aliases=network_aliases,
                              network_info=self.network_info,
                              volume_name=self._get_db_volume_name())
            database_info = \
                DatabaseInfo(host=db_ip_address, db_port=DB_PORT, bucketfs_port=BUCKETFS_PORT,
                             reused=reused, container_info=container_info)
            return database_info

    def _pull_docker_db_images_if_necessary(self):
        image_name = "exasol/docker-db"
        docker_db_image_info = ImageInfo(
            target_repository_name=image_name,
            source_repository_name=image_name,
            source_tag=self.docker_db_image_version,
            target_tag=self.docker_db_image_version,
            hash_value="", commit="",
            image_description=None)
        with self._get_docker_client() as docker_client:
            try:
                docker_client.images.get(docker_db_image_info.get_source_complete_name())
            except docker.errors.ImageNotFound as e:
                self.logger.info("Pulling docker-db image %s",
                                 docker_db_image_info.get_source_complete_name())
                output_generator = docker_client.api.pull(
                    docker_db_image_info.source_repository_name,
                    tag=docker_db_image_info.source_tag,
                    stream=True)
                self._handle_output(output_generator, docker_db_image_info)
        return docker_db_image_info

    def _prepare_db_volume(self, docker_client, db_private_network: str,
                           docker_db_image_info: ImageInfo) -> Volume:
        volume, container = self._prepare_volume(
            docker_client,
            self._get_db_volume_name(),
            self._get_db_volume_preparation_container_name(),
            remove_old_instances=True,
        )
        try:
            self._upload_init_db_files(container, db_private_network)
            self._execute_init_db(volume, container)
            return volume
        finally:
            container.remove(force=True)

    def _get_db_volume_preparation_container_name(self):
        return f"""{self.db_container_name}_preparation"""

    def _get_db_volume_name(self):
        return f"""{self.db_container_name}_volume"""

    def _remove_container(self, container_name:str):
        try:
            with self._get_docker_client() as docker_client:
                docker_client.containers.get(container_name).remove(force=True)
                self.logger.info("Removed container %s", container_name)
        except docker.errors.NotFound:
            pass

    def _remove_volume(self, volume_name:str):
        try:
            with self._get_docker_client() as docker_client:
                docker_client.volumes.get(volume_name).remove(force=True)
                self.logger.info("Removed volume %s", volume_name)
        except docker.errors.NotFound:
            pass

    def _prepare_volume(
            self,
            docker_client: docker.api.APIClient,
            volume_name,
            container_name,
            remove_old_instances: bool = False,
    ) -> Tuple[Volume, Container]:
        """
        Create an intermediate Docker Container containing a volume that
        can be mounted into another Docker Container.
        """
        if remove_old_instances:
            self._remove_container(container_name)
            self._remove_volume(volume_name)
        volume = docker_client.volumes.create(volume_name)
        container = docker_client.containers.run(
                image="ubuntu:18.04",
                name=container_name,
                auto_remove=True,
                command="sleep infinity",
                detach=True,
                volumes={volume.name: {"bind": "/exa", "mode": "rw"}},
                labels={
                    "test_environment_name": self.environment_name,
                    "container_type": "db_volume_preparation_container",
                }
        )
        return volume, container

    def _upload_init_db_files(self, container: Container, db_private_network: str):
        copy = DockerContainerCopy(container)
        init_db_script_str = pkg_resources.resource_string(
            PACKAGE_NAME,
            f"{self.docker_db_config_resource_name}/init_db.sh") # type: bytes

        copy.add_string_to_file("init_db.sh", init_db_script_str.decode("utf-8"))
        self._add_exa_conf(copy, db_private_network)
        copy.copy("/")

    def _add_exa_conf(self, copy: DockerContainerCopy,
                      db_private_network: str):
        certificate_dir = CERTIFICATES_MOUNT_DIR if self.certificate_volume_name is not None \
                            else CERTIFICATES_DEFAULT_DIR
        template_str = pkg_resources.resource_string(
            PACKAGE_NAME,
            f"{self.docker_db_config_resource_name}/EXAConf") # type: bytes
        template = Template(template_str.decode("utf-8"))
        additional_db_parameter_str = " ".join(self.additional_db_parameter)
        rendered_template = template.render(private_network=db_private_network,
                                            db_version=str(self.db_version),
                                            image_version=self.docker_db_image_version,
                                            mem_size=self.mem_size,
                                            disk_size=self.disk_size,
                                            name_servers=",".join(self.nameservers),
                                            certificate_dir=certificate_dir,
                                            additional_db_parameters=additional_db_parameter_str)
        copy.add_string_to_file("EXAConf", rendered_template)

    def _execute_init_db(self, db_volume: Volume, preparation_container: Container):
        disk_size_in_bytes = humanfriendly.parse_size(self.disk_size)
        min_overhead_in_gigabyte = 2  # Exasol needs at least a 2 GB larger device than the configured disk size
        overhead_factor = max(0.01, (
                min_overhead_in_gigabyte * 1024 * 1024 * 1024) / disk_size_in_bytes)  # and in general 1% larger
        device_size_in_bytes = disk_size_in_bytes * (1 + overhead_factor)
        device_size_in_megabytes = math.ceil(
            device_size_in_bytes / (1024 * 1024))  # The init_db.sh script works with MB, because its faster
        self.logger.info(
            f"Creating database volume of size {device_size_in_megabytes / 1024} GB using and overhead factor of {overhead_factor}")
        (exit_code, output) = preparation_container.exec_run(cmd=f"bash /init_db.sh {device_size_in_megabytes}")
        if exit_code != 0:
            raise Exception(
                "Error during preparation of docker-db volume %s got following output %s" % (db_volume.name, output))

    def cleanup_task(self, success):
        if (success and not self.no_database_cleanup_after_success) or \
                (not success and not self.no_database_cleanup_after_failure):
            volume_container = self._get_db_volume_preparation_container_name()
            try:
                self.logger.info(f"Cleaning up container %s", volume_container)
                self._remove_container(volume_container)
            except Exception as e:
                self.logger.error(f"Error during removing container %s: %s", volume_container, e)

            try:
                self.logger.info(f"Cleaning up container %s", self.db_container_name)
                self._remove_container(self.db_container_name)
            except Exception as e:
                self.logger.error(f"Error during removing container %s: %s", self.db_container_name, e)

            db_volume_name = self._get_db_volume_name()
            try:
                self.logger.info(f"Cleaning up docker volume %s", db_volume_name)
                self._remove_volume(db_volume_name)
            except Exception as e:
                self.logger.error(f"Error during removing docker volume %s: %s", db_volume_name, e)
